"""SAML 2.0 Utilities module for M2Crypto SSL functionality 

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "02/07/07"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: m2crypto.py 7322 2010-08-13 14:17:22Z pjkersha $'
import logging
log = logging.getLogger(__name__)

from warnings import warn # warn of impending certificate expiry

import os
import re

# Handle not before and not after strings
from time import strptime
from datetime import datetime

import M2Crypto
from M2Crypto import SSL, X509
from M2Crypto.httpslib import HTTPSConnection as _HTTPSConnection


class X500DNError(Exception):
    """Exception handling for NDG X.500 DN class."""


class X500DN(dict):
    "NDG X500 Distinguished name"
    
    # Class attribute - look-up mapping short name attributes to their long
    # name equivalents
    # * private *
    __shortNameLUT = {
        'commonName':               'CN',
        'organisationalUnitName':   'OU',
        'organisation':             'O',
        'countryName':              'C',
        'emailAddress':             'EMAILADDRESS',
        'localityName':             'L',
        'stateOrProvinceName':      'ST',
        'streetAddress':            'STREET',
        'domainComponent':          'DC',
        'userid':                   'UID'
    }
    SLASH_PARSER_RE_STR = '/(%s)=' % '|'.join(__shortNameLUT.keys() + 
                                              __shortNameLUT.values())    
    SLASH_PARSER_RE = re.compile(SLASH_PARSER_RE_STR)

    COMMA_PARSER_RE_STR = '[,]?\s*(%s)=' % '|'.join(__shortNameLUT.keys() + 
                                                    __shortNameLUT.values())    
    COMMA_PARSER_RE = re.compile(COMMA_PARSER_RE_STR)
    
    def __init__(self, dn=None, m2CryptoX509Name=None, separator=None):

        """Create a new X500 Distinguished Name

        @type m2CryptoX509Name: M2Crypto.X509.X509_Name
        @param m2CryptoX509Name:   initialise using using an 
        M2Crypto.X509.X509_Name
        @type dn: basestring
        @param dn: initialise using a distinguished name string
        @type separator: basestring
        @param: separator: separator used to delimit dn fields - usually '/' 
        or ','.  If dn is input and separator is omitted the separator 
        character will be automatically parsed from the dn string.
        """
        
        # Private key data
        self.__dat = {}.fromkeys(X500DN.__shortNameLUT.values(), '')
    
        dict.__init__(self)
    
        self.__separator = None
        
        # Check for separator from input
        if separator is not None:
            if not isinstance(separator, basestring):
                raise X500DNError("dn Separator must be a valid string")

            # Check for single character but allow trailing space chars
            if len(separator.lstrip()) is not 1:
                raise X500DNError("dn separator must be a single character")

            self.__separator = separator
        
        if m2CryptoX509Name is not None:
            # the argument is an x509 dn in m2crypto format
            self.deserialise(str(m2CryptoX509Name))
            
        elif dn is not None:
            # Separator can be parsed from the input DN string - only attempt
            # if no explict separator was input
            if self.__separator is None:
                self.__separator = self.parseSeparator(dn)
                
            # Split Distinguished name string into constituent fields
            self.deserialise(dn)

    @classmethod
    def fromString(cls, dn):
        """Convenience method for parsing DN string into a new instance
        """
        return cls(dn=dn)

    def __repr__(self):
        """Give representation based on underlying dict object"""
        return repr(self.__dat)
        
    def __str__(self):
        """Behaviour for print and string statements - convert DN into
        serialised format."""
        return self.serialise()
        
    def __eq__(self, x500dn):
        """Return true if the all the fields of the two DNs are equal"""
        
        if not isinstance(x500dn, X500DN):
            return False

        return self.__dat.items() == x500dn.items()
   
    def __ne__(self, x500dn):
        """Return true if the all the fields of the two DNs are equal"""
        
        if not isinstance(x500dn, X500DN):
            return False

        return self.__dat.items() != x500dn.items()
  
    def __delitem__(self, key):
        """Prevent keys from being deleted."""
        raise X500DNError('Keys cannot be deleted from the X500DN')

    def __getitem__(self, key):

        # Check input key
        if self.__dat.has_key(key):

            # key recognised
            return self.__dat[key]
        
        elif X500DN.__shortNameLUT.has_key(key):

            # key not recognised - but a long name version of the key may
            # have been passed
            shortName = X500DN.__shortNameLUT[key]
            return self.__dat[shortName]

        else:
            # key not recognised as a short or long name version
            raise KeyError('Key "' + key + '" not recognised for X500DN')

    def __setitem__(self, key, item):
        
        # Check input key
        if self.__dat.has_key(key):

            # key recognised
            self.__dat[key] = item
            
        elif X500DN.__shortNameLUT.has_key(key):
                
            # key not recognised - but a long name version of the key may
            # have been passed
            shortName = X500DN.__shortNameLUT[key]
            self.__dat[shortName] = item
            
        else:
            # key not recognised as a short or long name version
            raise KeyError('Key "' + key + '" not recognised for X500DN')

    def clear(self):
        raise X500DNError("Data cannot be cleared from X500DN")

    def copy(self):

        import copy
        return copy.copy(self)

    def keys(self):
        return self.__dat.keys()

    def items(self):
        return self.__dat.items()

    def values(self):
        return self.__dat.values()

    def has_key(self, key):
        return self.__dat.has_key(key)

    # 'in' operator
    def __contains__(self, key):
        return self.has_key(key)

    def get(self, *arg):
        return self.__dat.get(*arg)
  
    def serialise(self, separator=None):
        """Combine fields in Distinguished Name into a single string."""
        
        if separator:
            if not isinstance(separator, basestring):
                raise X500DNError("Separator must be a valid string")
        
            self.__separator = separator
            
        else:
            # Default to / if no separator is set
            separator = '/'


        # If using '/' then prepend DN with an initial '/' char
        if separator == '/':
            sDN = separator
        else:
            sDN = ''
     
        dnList = []
        for (key, val) in self.__dat.items():
            if val:
                if isinstance(val, tuple):
                    dnList += [separator.join(["%s=%s" % (key, valSub) \
                                               for valSub in val])]
                else:
                    dnList += ["%s=%s" % (key, val)]
                
        sDN += separator.join(dnList)
                                
        return sDN

    serialize = serialise
    
    def deserialise(self, dn, separator=None):
        """Break up a DN string into it's constituent fields and use to
        update the object's dictionary"""
        
        if separator:
            if not isinstance(separator, basestring):
                raise X500DNError("Separator must be a valid string")

            self.__separator = separator


        # If no separator has been set, parse if from the DN string            
        if self.__separator is None:
            self.__separator = self.parseSeparator(dn)

        if self.__separator == '/':
            parserRe = self.__class__.SLASH_PARSER_RE
            
        elif self.__separator == ',':
            parserRe = self.__class__.COMMA_PARSER_RE
        else:
            raise X500DNError("DN field separator %r not recognised" % 
                              self.__separator)
            
        try:
            dnFields = parserRe.split(dn)
            if len(dnFields) < 2:
                raise X500DNError("Error parsing DN string: \"%s\"" % dn)

            items = zip(dnFields[1::2], dnFields[2::2])
            
            # Reset existing dictionary values
            self.__dat.fromkeys(self.__dat, '')
            
            # Strip leading and trailing space chars and convert into a
            # dictionary
            parsedDN = {}
            for key, val in items:
                key = key.strip()
                if key in parsedDN:
                    if isinstance(parsedDN[key], tuple):
                        parsedDN[key] = tuple(list(parsedDN[key]) + [val])
                    else:
                        parsedDN[key] = (parsedDN[key], val)
                else:
                    parsedDN[key] = val
                
            # Copy matching DN fields
            for key, val in parsedDN.items():
                if key not in self.__dat and key not in self.__shortNameLUT:
                    raise X500DNError('Invalid field "%s" in input DN string' %
                                      key)

                self.__dat[key] = val

                
        except Exception, excep:
            raise X500DNError("Error de-serialising DN \"%s\": %s" % \
                              (dn, str(excep)))

    deserialize = deserialise
    
    def parseSeparator(self, dn):
        """Attempt to parse the separator character from a given input
        DN string.  If not found, return None

        DNs don't use standard separators e.g.

        /C=UK/O=eScience/OU=CLRC/L=DL/CN=AN Other
        CN=SUM Oneelse,L=Didcot, O=RAL,OU=SSTD

        This function isolates and identifies the character.  - In the above,
        '/' and ',' respectively"""


        # Make a regular expression containing all the possible field
        # identifiers with equal sign appended and 'or'ed together.  \W should
        # match the separator which preceeds the field name. \s* allows any
        # whitespace between field name and field separator to be taken into
        # account.
        #
        # The resulting match should be a list.  The first character in each
        # element in the list should be the field separator and should be the
        # same
        regExpr = '|'.join(['\W\s*'+i+'=' for i in self.__dat.keys()])
        match = re.findall(regExpr, dn)
            
        # In the first example above, the resulting match is:
        # ['/C=', '/O=', '/OU=', '/L=']
        # In each element the first character is the separator
        sepList = [i[0:1] for i in match]

        # All separators should be the same character - return None if they
        # don't match
        if not [i for i in sepList if i != sepList[0]]:
            return sepList[0]
        else:
            return None

    @classmethod
    def Parse(cls, dn):
        """Convenience method to create an X500DN object from a DN string
        @type dn: basestring
        @param dn: Distinguished Name 
        """
        return cls(dn=dn)
    
    Deserialise = Deserialize = Parse


class X509CertError(Exception):
    """Exception handling for NDG X.509 Certificate handling class."""

class X509CertReadError(X509CertError):
    """Error reading in certificate from file"""

class X509CertParseError(X509CertError):
    """Error parsing a certificate"""
  
class X509CertInvalidNotBeforeTime(X509CertError):
    """Call from X509Cert.isValidTime if certificates not before time is
    BEFORE the current system time"""
    
class X509CertExpired(X509CertError):
    """Call from X509Cert.isValidTime if certificate has expired"""
    
   
class X509Cert(object):
    "NDG X509 Certificate Handling"

    formatPEM = M2Crypto.X509.FORMAT_PEM
    formatDER = M2Crypto.X509.FORMAT_DER
    
    def __init__(self, filePath=None, m2CryptoX509=None):

        # Set certificate file path
        if filePath is not None:
            if not isinstance(filePath, basestring):
                raise X509CertError("Certificate File Path input must be a "
                                    "valid string")
            
        self.__filePath = filePath            
        self.__dn = None
        self.__dtNotBefore = None
        self.__dtNotAfter = None
        
        if m2CryptoX509:
            self.__setM2CryptoX509(m2CryptoX509)
        else:
            self.__m2CryptoX509 = None

    def read(self, 
             filePath=None, 
             format=None, 
             warningStackLevel=3,
             **isValidTimeKw):
        """Read a certificate from PEM encoded DER format file
        
        @type filePath: basestring
        @param filePath: file path of PEM format file to be read
        
        @type format: int
        @param format: format of input file - PEM is the default.  Set to
        X509Cert.formatDER for DER format
        
        @type isValidTimeKw: dict
        @param isValidTimeKw: keywords to isValidTime() call"""

        if format is None:
            format = X509Cert.formatPEM
        
         # Check for optional input certificate file path
        if filePath is not None:
            if not isinstance(filePath, basestring):
                raise X509CertError("Certificate File Path input must be a "
                                    "valid string")
            
            self.__filePath = filePath
       
        try:
            self.__m2CryptoX509 = M2Crypto.X509.load_cert(self.__filePath,
                                                          format=format)
        except Exception, e:
            raise X509CertReadError("Error loading certificate \"%s\": %s" %
                                    (self.__filePath, e))

        # Update DN and validity times from M2Crypto X509 object just
        # created
        self.__setM2CryptoX509()
        
        self.isValidTime(warningStackLevel=warningStackLevel, **isValidTimeKw)

    def parse(self, 
              certTxt, 
              format=None, 
              warningStackLevel=3,
              **isValidTimeKw):
        """Read a certificate input as a string
        
        @type certTxt: basestring
        @param certTxt: PEM encoded certificate to parse 
        
        @type format: int
        @param format: format of input file - PEM is the default.  Set to
        X509Cert.formatDER for DER format
        
        @type isValidTimeKw: dict
        @param isValidTimeKw: keywords to isValidTime() call"""

        if format is None:
            format = X509Cert.formatPEM
            
        try:
            # Create M2Crypto memory buffer and pass to load certificate
            # method
            #
            # Nb. input converted to standard string - buffer method won't
            # accept unicode type strings
#            certBIO = M2Crypto.BIO.MemoryBuffer(str(certTxt))
#            self.__m2CryptoX509 = M2Crypto.X509.load_cert_bio(certBIO)
            self.__m2CryptoX509 = M2Crypto.X509.load_cert_string(str(certTxt),
                                                                 format=format)
        except Exception, e:
            raise X509CertParseError("Error loading certificate: %s" % e)

        # Update DN and validity times from M2Crypto X509 object just
        # created
        self.__setM2CryptoX509()
        
        self.isValidTime(warningStackLevel=warningStackLevel, **isValidTimeKw)
      
    def __setM2CryptoX509(self, m2CryptoX509=None):
        """Private method allows class members to be updated from the
        current M2Crypto object.  __m2CryptoX509 must have been set."""
        
        if m2CryptoX509 is not None:
            if not isinstance(m2CryptoX509, M2Crypto.X509.X509):
                raise TypeError("Incorrect type for input M2Crypto.X509.X509 "
                                "object")
                    
            self.__m2CryptoX509 = m2CryptoX509
                   
        # Get distinguished name
        m2CryptoX509Name = self.__m2CryptoX509.get_subject()

        # Instantiate X500 Distinguished name
        self.__dn = X500DN(m2CryptoX509Name=m2CryptoX509Name)
        
        # Get not before and not after validity times
        #
        # Only option for M2Crypto seems to be to return the times as
        # formatted strings and then parse them in order to create a datetime
        # type
        
        try:
            m2CryptoNotBefore = self.__m2CryptoX509.get_not_before()
            self.__dtNotBefore=self.__m2CryptoUTC2datetime(m2CryptoNotBefore)
                                        
        except Exception, e:
            raise X509CertError("Not Before time: %s" % e)

        try:
            m2CryptoNotAfter = self.__m2CryptoX509.get_not_after()
            self.__dtNotAfter = self.__m2CryptoUTC2datetime(m2CryptoNotAfter)
                                    
        except Exception, e:
            raise X509CertError("Not After time: %s" % e)

    def __getM2CryptoX509(self, m2CryptoX509=None):
        "Return M2Crypto X.509 cert object"
        return self.__m2CryptoX509
    
    m2CryptoX509 = property(fset=__setM2CryptoX509,
                            fget=__getM2CryptoX509,
                            doc="M2Crypto.X509.X509 type")
        
    def toString(self, **kw):
        """Return certificate file content as a PEM format 
        string"""
        return self.asPEM(**kw)
        
    def asPEM(self, filePath=None):
        """Return certificate file content as a PEM format 
        string"""
        
        # Check M2Crypto.X509 object has been instantiated - if not call
        # read method
        if self.__m2CryptoX509 is None:
            self.read(filePath)
            
        return self.__m2CryptoX509.as_pem()

    def asDER(self):
        """Return certificate file content in DER format"""
        
        # Check M2Crypto.X509 object has been instantiated 
        assert(self.__m2CryptoX509)
        return self.__m2CryptoX509.as_der()

    # Make some attributes accessible as read-only
    def __getDN(self):
        """Get X500 Distinguished Name."""
        return self.__dn

    dn = property(fget=__getDN, doc="X.509 Distinguished Name")
    
    def __getVersion(self):
        """Get X.509 Certificate version"""
        if self.__m2CryptoX509 is None:
            return None
        
        return self.__m2CryptoX509.get_version()

    version = property(fget=__getVersion, doc="X.509 Certificate version")
    
    def __getSerialNumber(self):
        """Get Serial Number"""
        if self.__m2CryptoX509 is None:
            return None
        
        return self.__m2CryptoX509.get_serial_number()
    
    serialNumber = property(fget=__getSerialNumber, 
                            doc="X.509 Certificate Serial Number")

    def __getNotBefore(self):
        """Get not before validity time as datetime type"""
        if self.__m2CryptoX509 is None:
            return None
        
        return self.__dtNotBefore

    notBefore = property(fget=__getNotBefore, 
                         doc="Not before validity time as datetime type")
    
    def __getNotAfter(self):
        """Get not after validity time as datetime type"""
        if self.__m2CryptoX509 is None:
            return None
        
        return self.__dtNotAfter

    notAfter = property(fget=__getNotAfter, 
                         doc="Not after validity time as datetime type")
    
    def __getPubKey(self):
        """Get public key
        
        @return: RSA public key for certificate
        @rtype: M2Crypto.RSA.RSA_pub"""
        if self.__m2CryptoX509 is None:
            return None
        
        return self.__m2CryptoX509.get_pubkey()

    pubKey = property(fget=__getPubKey, doc="Public Key")
    
    def __getIssuer(self):
        """Get Certificate issuer"""
        if self.__m2CryptoX509 is None:
            return None
        
        # Return as X500DN type
        return X500DN(m2CryptoX509Name=self.__m2CryptoX509.get_issuer())

    issuer = property(fget=__getIssuer, doc="Certificate Issuer")
    
    def __getSubject(self):
        """Get Certificate subject"""
        if self.__m2CryptoX509 is None:
            return None

        # Return as X500DN type
        return X500DN(m2CryptoX509Name=self.__m2CryptoX509.get_subject())
    
    subject = property(fget=__getSubject, doc="Certificate subject")

    def isValidTime(self, 
                    raiseExcep=False, 
                    expiryWarning=True, 
                    nDaysBeforeExpiryLimit=30,
                    warningStackLevel=2):
        """Check Certificate for expiry

        @type raiseExcep: bool
        @param raiseExcep: set True to raise an exception if certificate is 
        invalid
        
        @type expiryWarning: bool
        @param expiryWarning: set to True to output a warning message if the 
        certificate is due to expire in less than nDaysBeforeExpiryLimit days. 
        Message is sent using warnings.warn and through logging.warning.  No 
        message is set if the certificate has an otherwise invalid time
        
        @type nDaysBeforeExpiryLimit: int
        @param nDaysBeforeExpiryLimit: used in conjunction with the 
        expiryWarning flag.  Set the number of days in advance of certificate
        expiry from which to start outputing warnings
        
        @type warningStackLevel: int
        @param warningStackLevel: set where in the stack to flag the warning
        from.  Level 2 will flag it at the level of the caller of this 
        method.  Level 3 would flag at the level of the caller of the caller
        and so on.
        
        @raise X509CertInvalidNotBeforeTime: current time is before the 
        certificate's notBefore time
        @raise X509CertExpired: current time is after the certificate's 
        notAfter time"""

        if not isinstance(self.__dtNotBefore, datetime):
            raise X509CertError("Not Before datetime is not set")

        if not isinstance(self.__dtNotAfter, datetime):
            raise X509CertError("Not After datetime is not set")
       
        dtNow = datetime.utcnow()
        isValidTime = dtNow > self.__dtNotBefore and dtNow < self.__dtNotAfter

        # Helper string for message output
        if self.__filePath:
            fileInfo = ' "%s"' % self.__filePath
        else:
            fileInfo = ''
             
        
        # Set a warning message for impending expiry of certificate but only
        # if the certificate is not any other way invalid - see below
        if isValidTime and expiryWarning:
            dtTime2Expiry = self.__dtNotAfter - dtNow
            if dtTime2Expiry.days < nDaysBeforeExpiryLimit:
                msg = ('Certificate%s with DN "%s" will expire in %d days on: '
                       '%s' % (fileInfo, 
                               self.dn, 
                               dtTime2Expiry.days, 
                               self.__dtNotAfter))
                warn(msg, stacklevel=warningStackLevel)
                log.warning(msg)
        
                     
        if dtNow < self.__dtNotBefore:
            msg = ("Current time %s is before the certificate's Not Before "
                   'Time %s for certificate%s with DN "%s"' % 
                   (dtNow, self.__dtNotBefore, fileInfo, self.dn))
            log.error(msg)
            if raiseExcep:
                raise X509CertInvalidNotBeforeTime(msg)
            
        elif dtNow > self.__dtNotAfter:
            msg = ('Certificate%s with DN "%s" has expired: the time now is '
                   '%s and the certificate expiry is %s.' %(fileInfo,
                                                            self.dn, 
                                                            dtNow, 
                                                            self.__dtNotAfter))
            log.error(msg)
            if raiseExcep:
                raise X509CertExpired(msg)

        # If exception flag is not set return validity as bool
        return isValidTime

    def __m2CryptoUTC2datetime(self, m2CryptoUTC):
        """Convert M2Crypto UTC time string as returned by get_not_before/
        get_not_after methods into datetime type"""
        
        datetimeRE = "([a-zA-Z]{3} {1,2}\d{1,2} \d{2}:\d{2}:\d{2} \d{4}).*"
        sM2CryptoUTC = None
        
        try:
            # Convert into string
            sM2CryptoUTC = str(m2CryptoUTC)
            
            # Check for expected format - string may have trailing GMT - ignore
            sTime = re.findall(datetimeRE, sM2CryptoUTC)[0]

            # Convert into a tuple
            lTime = strptime(sTime, "%b %d %H:%M:%S %Y")[0:6]

            return datetime(lTime[0], lTime[1], lTime[2],
                            lTime[3], lTime[4], lTime[5])
                                    
        except Exception, e:
            msg = "Error parsing M2Crypto UTC"
            if sM2CryptoUTC is not None:
                msg += ": " + sM2CryptoUTC
                
            raise X509CertError(msg)
    
    def verify(self, pubKey, **kw):
        """Verify a certificate against the public key of the
        issuer
        
        @param pubKey: public key of cert that issued self
        @type pubKey: M2Crypto.RSA.RSA_pub
        @param **kw: keywords to pass to M2Crypto.X509.X509 -
        'pkey'
        @type: dict
        @return: True if verifies OK, False otherwise
        @rtype: bool
        """
        return bool(self.__m2CryptoX509.verify(pubKey, **kw))

    @classmethod
    def Read(cls, filePath, warningStackLevel=4, **isValidTimeKw):
        """Create a new X509 certificate read in from a file"""
        x509Cert = cls(filePath=filePath)  
        x509Cert.read(warningStackLevel=warningStackLevel, **isValidTimeKw)
        
        return x509Cert
    
    @classmethod
    def Parse(cls, x509CertTxt, warningStackLevel=4, **isValidTimeKw):
        """Create a new X509 certificate from string of file content"""
        x509Cert = cls()      
        x509Cert.parse(x509CertTxt, 
                       warningStackLevel=warningStackLevel,
                       **isValidTimeKw)
        
        return x509Cert

    @classmethod
    def fromM2Crypto(cls, m2CryptoX509):
        """Convenience method to instantiate a new object from an M2Crypto
        X.509 certificate object"""
        x509Cert = cls(m2CryptoX509=m2CryptoX509)
        return x509Cert
    
    
class X509StackError(X509CertError):
    """Error from X509Stack type"""


class X509StackEmptyError(X509CertError):
    """Expecting non-zero length X509Stack"""


class X509CertIssuerNotFound(X509CertError):
    """Raise from verifyCertChain if no certificate can be found to verify the
    input"""


class SelfSignedCert(X509CertError):
    """Raise from verifyCertChain if cert. is self-signed and 
    rejectSelfSignedCert=True"""


class X509CertInvalidSignature(X509CertError):
    """X.509 Certificate has an invalid signature"""       
       
class X509Stack(object):
    """Wrapper for M2Crypto X509_Stack"""
    
    def __init__(self, m2X509Stack=None):
        """Initialise from an M2Crypto stack object
        
        @param m2X509Stack: M2Crypto X.509 stack object
        @type m2X509Stack: M2Crypto.X509.X509_Stack"""
        
        self.__m2X509Stack = m2X509Stack or M2Crypto.X509.X509_Stack()
        
    def __len__(self):
        """@return: length of stack
        @rtype: int"""
        return self.__m2X509Stack.__len__()

    def __getitem__(self, idx):
        """Index stack as an array
        @param idx: stack index
        @type idx: int
        @return: X.509 cert object
        @rtype: ndg.security.common.X509.X509Cert"""
        
        return X509Cert(m2CryptoX509=self.__m2X509Stack.__getitem__(idx))
    
    def __iter__(self):
        """@return: stack iterator
        @rtype: listiterator"""
        return iter([X509Cert(m2CryptoX509=i) for i in self.__m2X509Stack])

    def push(self, x509Cert):
        """Push an X509 certificate onto the stack.
        
        @param x509Cert: X509 object.
        @type x509Cert: M2Crypto.X509.X509,
        ndg.security.common.X509.X509Cert or basestring
        @return: The number of X509 objects currently on the stack.
        @rtype: int"""
        if isinstance(x509Cert, M2Crypto.X509.X509):
            return self.__m2X509Stack.push(x509Cert)
        
        elif isinstance(x509Cert, X509Cert):
            return self.__m2X509Stack.push(x509Cert.m2CryptoX509)
        
        elif isinstance(x509Cert, basestring):
            return self.__m2X509Stack.push(\
                                       X509Cert.Parse(x509Cert).m2CryptoX509)            
        else:
            raise X509StackError("Expecting M2Crypto.X509.X509, ndg.security."
                                 "common.X509.X509Cert or string type")
                
    def pop(self):
        """Pop a certificate from the stack.
        
        @return: X509 object that was popped, or None if there is nothing
        to pop.
        @rtype: ndg.security.common.X509.X509Cert
        """
        return X509Cert(m2CryptoX509=self.__m2X509Stack.pop())

    def asDER(self):
        """Return the stack as a DER encoded string
        @return: DER string
        @rtype: string"""
        return self.__m2X509Stack.as_der()

    def verifyCertChain(self, 
                        x509Cert2Verify=None, 
                        caX509Stack=[],
                        rejectSelfSignedCert=True):
        """Treat stack as a list of certificates in a chain of
        trust.  Validate the signatures through to a single root issuer.  

        @param x509Cert2Verify: X.509 certificate to be verified default is
        last in the stack
        @type x509Cert2Verify: X509Cert
        
        @param caX509Stack: X.509 stack containing CA certificates that are
        trusted.
        @type caX509Stack: X509Stack
        
        @param rejectSelfSignedCert: Set to True (default) to raise an 
        SelfSignedCert exception if a certificate in self's stack is 
        self-signed.  
        @type rejectSelfSignedCert: bool"""
        
        n2Validate = len(self)
        if x509Cert2Verify:
            # One more to validate in addition to stack content
            n2Validate += 1
        else:
            # Validate starting from last on stack - but check first that it's
            # populated
            if n2Validate == 0:
                raise X509StackEmptyError("Empty stack and no x509Cert2Verify "
                                          "set: no cert.s to verify")

            x509Cert2Verify = self[-1]
              
        # Exit loop if all certs have been validated or if find a self 
        # signed cert.
        nValidated = 0
        issuerX509Cert = None
        while nValidated < n2Validate:
            issuerX509Cert = None
            issuerDN = x509Cert2Verify.issuer
            
            # Search for issuing certificate in stack
            for x509Cert in self:
                if x509Cert.dn == issuerDN:
                    # Match found - the cert.'s issuer has been found in the 
                    # stack
                    issuerX509Cert = x509Cert
                    break
                    
            if issuerX509Cert:
                # An issuing cert. has been found - use it to check the 
                # signature of the cert. to be verified
                if not x509Cert2Verify.verify(issuerX509Cert.pubKey):
                    X509CertInvalidSignature('Signature is invalid for cert. '
                                             '"%s"' % x509Cert2Verify.dn)
                
                # In the next iteration the issuer cert. will be checked:
                # 1) search for a cert. in the stack that issued it
                # 2) If found use the issuing cert. to verify
                x509Cert2Verify = issuerX509Cert
                nValidated += 1
            else:
                # All certs in the stack have been searched
                break


        if issuerX509Cert:            
            # Check for self-signed certificate
            if (nValidated == 1 and rejectSelfSignedCert and 
                issuerX509Cert.dn == issuerX509Cert.issuer):

                # If only one iteration occurred then it must be a self
                # signed certificate
                raise SelfSignedCert("Certificate is self signed: [DN=%s]" %
                                     issuerX509Cert.dn)
           
            if not caX509Stack:
                caX509Stack = [issuerX509Cert]
                         
        elif not caX509Stack:
            raise X509CertIssuerNotFound('No issuer certificate found for '
                                         'certificate "%s"' % 
                                         x509Cert2Verify.dn)
            
        for caCert in caX509Stack:
            issuerDN = x509Cert2Verify.issuer
            if caCert.dn == issuerDN:
                issuerX509Cert = caCert
                break
        
        if issuerX509Cert:
            if not x509Cert2Verify.verify(issuerX509Cert.pubKey):
                X509CertInvalidSignature('Signature is invalid for cert. "%s"' %
                                         x509Cert2Verify.dn)
            
            # Chain is validated through to CA cert
            return
        else:
            raise X509CertIssuerNotFound('No issuer cert. found for '
                                         'certificate "%s"'%x509Cert2Verify.dn)
        
        # If this point is reached then an issuing cert is missing from the
        # chain        
        raise X509CertIssuerNotFound('Can\'t find issuer cert "%s" for '
                                     'certificate "%s"' %
                                     (x509Cert2Verify.issuer, 
                                      x509Cert2Verify.dn))


class InvalidCertSignature(SSL.Checker.SSLVerificationError):
    """Raise if verification against CA cert public key fails"""


class InvalidCertDN(SSL.Checker.SSLVerificationError):
    """Raise if verification against a list acceptable DNs fails"""
   

class HostCheck(SSL.Checker.Checker, object):
    """Override SSL.Checker.Checker to enable alternate Common Name
    setting match for peer cert"""

    def __init__(self, 
                 peerCertDN=None, 
                 peerCertCN=None,
                 acceptedDNs=[], 
                 caCertList=[],
                 caCertFilePathList=[], 
                 **kw):
        """Override parent class __init__ to enable setting of myProxyServerDN
        setting
        
        @type peerCertDN: string/list
        @param peerCertDN: Set the expected Distinguished Name of the
        server to avoid errors matching hostnames.  This is useful
        where the hostname is not fully qualified.  

        *param acceptedDNs: a list of acceptable DNs.  This enables validation 
        where the expected DN is where against a limited list of certs.
        
        @type peerCertCN: string
        @param peerCertCN: enable alternate Common Name to peer
        hostname
        
        @type caCertList: list type of M2Crypto.X509.X509 types
        @param caCertList: CA X.509 certificates - if set the peer cert's 
        CA signature is verified against one of these.  At least one must
        verify
        
        @type caCertFilePathList: list string types
        @param caCertFilePathList: same as caCertList except input as list
        of CA cert file paths"""
        
        SSL.Checker.Checker.__init__(self, **kw)
        
        self.peerCertDN = peerCertDN
        self.peerCertCN = peerCertCN
        self.acceptedDNs = acceptedDNs
        
        if caCertList:
            self.caCertList = caCertList
        elif caCertFilePathList:
            self.caCertFilePathList = caCertFilePathList
        else:
            # Set default to enable len() test in __call__
            self.__caCertStack = ()
            
    def __call__(self, peerCert, host=None):
        """Carry out checks on server ID
        @param peerCert: MyProxy server host certificate as M2Crypto.X509.X509
        instance
        @param host: name of host to check
        """
        if peerCert is None:
            raise SSL.Checker.NoCertificate('SSL Peer did not return '
                                            'certificate')

        peerCertDN = '/'+peerCert.get_subject().as_text().replace(', ', '/')
        try:
            SSL.Checker.Checker.__call__(self, peerCert, host=self.peerCertCN)
            
        except SSL.Checker.WrongHost, e:
            # Try match against peerCertDN set   
            if peerCertDN != self.peerCertDN:
                raise e

        # At least one match should be found in the list - first convert to
        # NDG X500DN type to allow per field matching for DN comparison
        peerCertX500DN = X500DN(dn=peerCertDN)
        
        if self.acceptedDNs:
           matchFound = False
           for dn in self.acceptedDNs:
               x500dn = X500DN(dn=dn)
               if x500dn == peerCertX500DN:
                   matchFound = True
                   break
               
           if not matchFound:
               raise InvalidCertDN('Peer cert DN "%s" doesn\'t match '
                                   'verification list' % peerCertDN)

        if len(self.__caCertStack) > 0:
            try:
                self.__caCertStack.verifyCertChain(
                           x509Cert2Verify=X509Cert(m2CryptoX509=peerCert))
            except Exception, e:
                raise InvalidCertSignature("Peer certificate verification "
                                           "against CA certificate failed: %s" 
                                           % e)
              
        # They match - drop the exception and return all OK instead          
        return True
      
    def __setCACertList(self, caCertList):
        """Set list of CA certs - peer cert must validate against at least one
        of these"""
        self.__caCertStack = X509Stack()
        for caCert in caCertList:
            self.__caCertStack.push(caCert)

    caCertList = property(fset=__setCACertList,
                          doc="list of CA certificates - the peer certificate "
                              "must validate against one")

    def __setCACertsFromFileList(self, caCertFilePathList):
        '''Read CA certificates from file and add them to the X.509
        stack
        
        @type caCertFilePathList: basestring, list or tuple
        @param caCertFilePathList: list of file paths for CA certificates to
        be used to verify certificate used to sign message.  If a single 
        string item is input then this is converted into a tuple
        '''
        if isinstance(caCertFilePathList, basestring):
            caCertFilePathList = (caCertFilePathList,)
            
        elif not isinstance(caCertFilePathList, (list, tuple)):
            raise TypeError('Expecting a basestring, list or tuple type for '
                            '"caCertFilePathList"')

        self.__caCertStack = X509Stack()

        for caCertFilePath in caCertFilePathList:
            self.__caCertStack.push(X509.load_cert(caCertFilePath))
        
    caCertFilePathList = property(fset=__setCACertsFromFileList,
                                  doc="list of CA certificate file paths - "
                                      "peer certificate must validate against "
                                      "one")


class HTTPSConnection(_HTTPSConnection):
    """Modified version of M2Crypto equivalent to enable custom checks with
    the peer and timeout settings
    
    @type defReadTimeout: M2Crypto.SSL.timeout
    @cvar defReadTimeout: default timeout for read operations
    @type defWriteTimeout: M2Crypto.SSL.timeout
    @cvar defWriteTimeout: default timeout for write operations"""    
    defReadTimeout = SSL.timeout(sec=20.)
    defWriteTimeout = SSL.timeout(sec=20.)
    
    def __init__(self, *args, **kw):
        '''Overload to enable setting of post connection check
        callback to SSL.Connection
        
        type *args: tuple
        param *args: args which apply to M2Crypto.httpslib.HTTPSConnection
        type **kw: dict
        param **kw: additional keywords
        @type postConnectionCheck: SSL.Checker.Checker derivative
        @keyword postConnectionCheck: set class for checking peer
        @type readTimeout: M2Crypto.SSL.timeout
        @keyword readTimeout: readTimeout - set timeout for read
        @type writeTimeout: M2Crypto.SSL.timeout
        @keyword writeTimeout: similar to read timeout'''
        
        self._postConnectionCheck = kw.pop('postConnectionCheck',
                                           SSL.Checker.Checker)
        
        if 'readTimeout' in kw:
            if not isinstance(kw['readTimeout'], SSL.timeout):
                raise AttributeError("readTimeout must be of type "
                                     "M2Crypto.SSL.timeout")
            self.readTimeout = kw.pop('readTimeout')
        else:
            self.readTimeout = HTTPSConnection.defReadTimeout
              
        if 'writeTimeout' in kw:
            if not isinstance(kw['writeTimeout'], SSL.timeout):
                raise AttributeError("writeTimeout must be of type "
                                     "M2Crypto.SSL.timeout") 
            self.writeTimeout = kw.pop('writeTimeout')
        else:
            self.writeTimeout = HTTPSConnection.defWriteTimeout
    
        self._clntCertFilePath = kw.pop('clntCertFilePath', None)
        self._clntPriKeyFilePath = kw.pop('clntPriKeyFilePath', None)
        
        _HTTPSConnection.__init__(self, *args, **kw)
        
        # load up certificate stuff
        if (self._clntCertFilePath is not None and 
            self._clntPriKeyFilePath is not None):
            self.ssl_ctx.load_cert(self._clntCertFilePath, 
                                   self._clntPriKeyFilePath)
        
        
    def connect(self):
        '''Overload M2Crypto.httpslib.HTTPSConnection to enable
        custom post connection check of peer certificate and socket timeout'''

        self.sock = SSL.Connection(self.ssl_ctx)
        self.sock.set_post_connection_check_callback(self._postConnectionCheck)

        self.sock.set_socket_read_timeout(self.readTimeout)
        self.sock.set_socket_write_timeout(self.writeTimeout)

        self.sock.connect((self.host, self.port))

    def putrequest(self, method, url, **kw):
        '''Overload to work around bug with unicode type URL'''
        url = str(url)
        _HTTPSConnection.putrequest(self, method, url, **kw) 
         
              
class SSLContextProxy(object):
    """Holder for M2Crypto.SSL.Context parameters"""
    PRE_VERIFY_FAIL, PRE_VERIFY_OK = range(2)
    
    SSL_CERT_FILEPATH_OPTNAME = "sslCertFilePath"
    SSL_PRIKEY_FILEPATH_OPTNAME = "sslPriKeyFilePath"
    SSL_PRIKEY_PWD_OPTNAME = "sslPriKeyPwd"
    SSL_CACERT_FILEPATH_OPTNAME = "sslCACertFilePath"
    SSL_CACERT_DIRPATH_OPTNAME = "sslCACertDir"
    SSL_VALID_DNS_OPTNAME = "sslValidDNs"
    
    OPTNAMES = (
        SSL_CERT_FILEPATH_OPTNAME,
        SSL_PRIKEY_FILEPATH_OPTNAME,
        SSL_PRIKEY_PWD_OPTNAME,
        SSL_CACERT_FILEPATH_OPTNAME,
        SSL_CACERT_DIRPATH_OPTNAME,
        SSL_VALID_DNS_OPTNAME
    )
    
    __slots__ = tuple(["__%s" % name for name in OPTNAMES])
    del name
    
    VALID_DNS_PAT = re.compile(',\s*')
    
    def __init__(self):
        self.__sslCertFilePath = None
        self.__sslPriKeyFilePath = None
        self.__sslPriKeyPwd = None
        self.__sslCACertFilePath = None
        self.__sslCACertDir = None
        self.__sslValidDNs = []

    def createCtx(self, depth=9, **kw):
        """Create an M2Crypto SSL Context from this objects properties
        @type depth: int
        @param depth: max. depth of certificate to verify against
        @type kw: dict
        @param kw: M2Crypto.SSL.Context keyword arguments
        @rtype: M2Crypto.SSL.Context
        @return M2Crypto SSL context object
        """
        ctx = SSL.Context(**kw)
        
        # Configure context according to this proxy's attributes
        if self.sslCertFilePath and self.sslPriKeyFilePath:
            # Pass client certificate
            ctx.load_cert(self.sslCertFilePath, 
                          self.__sslPriKeyFilePath, 
                          lambda *arg, **kw: self.sslPriKeyPwd)
            log.debug("Set client certificate and key in SSL Context")
        else:
            log.debug("No client certificate or key set in SSL Context")
            
        if self.sslCACertFilePath or self.sslCACertDir:
            # Set CA certificates in order to verify peer
            ctx.load_verify_locations(self.sslCACertFilePath, 
                                      self.sslCACertDir)
            mode = SSL.verify_peer|SSL.verify_fail_if_no_peer_cert
        else:
            mode = SSL.verify_fail_if_no_peer_cert
            log.warning('No CA certificate files set: mode set to '
                        '"verify_fail_if_no_peer_cert" only')
            
        if len(self.sslValidDNs) > 0:
            # Set custom callback in order to verify peer certificate DN 
            # against whitelist
            callback = self.createVerifySSLPeerCertCallback()
            log.debug('Set peer certificate Distinguished Name check set in '
                      'SSL Context')
        else:
            callback = None
            log.warning('No peer certificate Distinguished Name check set in '
                        'SSL Context')
            
        ctx.set_verify(mode, depth, callback=callback)
           
        return ctx
 
    def copy(self, sslCtxProxy):
        """Copy settings from another context object
        """
        if not isinstance(sslCtxProxy, SSLContextProxy):
            raise TypeError('Expecting %r for copy method input object; '
                            'got %r' % (SSLContextProxy, type(sslCtxProxy)))
        
        for name in SSLContextProxy.OPTNAMES:
            setattr(self, name, getattr(sslCtxProxy, name))
            
    def createVerifySSLPeerCertCallback(self):
        """Create a callback function to enable the DN of the peer in an SSL
        connection to be verified against a whitelist.  
        
        Nb. Making this function within the scope of a method of the class to
        enables to access instance variables
        """
        
        def _verifySSLPeerCertCallback(preVerifyOK, x509StoreCtx):
            '''SSL verify callback function used to control the behaviour when 
            the SSL_VERIFY_PEER flag is set.  See:
            
            http://www.openssl.org/docs/ssl/SSL_CTX_set_verify.html
            
            This implementation applies verification in order to check the DN
            of the peer certificate against a whitelist
            
            @type preVerifyOK: int
            @param preVerifyOK: If a verification error is found, this 
            parameter will be set to 0
            @type x509StoreCtx: M2Crypto.X509.X509_Store_Context
            @param x509StoreCtx: locate the certificate to be verified and 
            perform additional verification steps as needed
            @rtype: int
            @return: controls the strategy of the further verification process. 
            - If verify_callback returns 0, the verification process is 
            immediately stopped with "verification failed" state. If 
            SSL_VERIFY_PEER is set, a verification failure alert is sent to the
            peer and the TLS/SSL handshake is terminated. 
            - If verify_callback returns 1, the verification process is 
            continued. 
            If verify_callback always returns 1, the TLS/SSL handshake will not
            be terminated with respect to verification failures and the 
            connection 
            will be established. The calling process can however retrieve the 
            error code of the last verification error using 
            SSL_get_verify_result or by maintaining its own error storage 
            managed by verify_callback.
            '''
            if preVerifyOK == 0:
                # Something is wrong with the certificate don't bother 
                # proceeding any further
                log.error("verifyCallback: pre-verify OK flagged an error "
                          "with the peer certificate, returning error state "
                          "to caller ...")
                return preVerifyOK
            
            x509CertChain = x509StoreCtx.get1_chain()
            for cert in x509CertChain:
                x509Cert = X509Cert.fromM2Crypto(cert)
                if x509Cert.dn in self.sslValidDNs:
                    return preVerifyOK
                
                subject = cert.get_subject()
                dn = subject.as_text()
                log.debug("verifyCallback: dn = %r", dn)
                
            # No match found so return fail status
            log.debug("No match for peer certificate %s in DN whitelist %r",
                      x509Cert.dn, self.sslValidDNs)
            return SSLContextProxy.PRE_VERIFY_FAIL
        
        return _verifySSLPeerCertCallback

    def _getSSLCertFilePath(self):
        return self.__sslCertFilePath
    
    def _setSSLCertFilePath(self, filePath):
        "Set X.509 cert file path property method"
        
        if isinstance(filePath, basestring):
            filePath = os.path.expandvars(filePath)
            
        elif filePath is not None:
            raise TypeError("X.509 cert. file path must be a valid string")
        
        self.__sslCertFilePath = filePath
                
    sslCertFilePath = property(fset=_setSSLCertFilePath,
                               fget=_getSSLCertFilePath,
                               doc="File path to X.509 cert.")
        
    def _getSSLCACertFilePath(self):
        """Get file path for list of CA cert or certs used to validate SSL 
        connections
        
        @rtype sslCACertFilePath: basestring
        @return sslCACertFilePathList: file path to file containing concatenated
        PEM encoded CA certificates."""
        return self.__sslCACertFilePath
    
    def _setSSLCACertFilePath(self, value):
        """Set CA cert file path
        
        @type sslCACertFilePath: basestring, list, tuple or None
        @param sslCACertFilePath: file path to CA certificate file.  If None
        then the input is quietly ignored."""
        if isinstance(value, basestring):
            self.__sslCACertFilePath = os.path.expandvars(value)
            
        elif value is None:
            self.__sslCACertFilePath = value
            
        else:
            raise TypeError("Input CA Certificate file path must be "
                            "a valid string or None type: %r" % type(value)) 
        
        
    sslCACertFilePath = property(fget=_getSSLCACertFilePath,
                                 fset=_setSSLCACertFilePath,
                                 doc="Path to file containing concatenated PEM "
                                     "encoded CA Certificates - used for "
                                     "verification of peer certs in SSL "
                                     "connection")
       
    def _getSSLCACertDir(self):
        """Get file path for list of CA cert or certs used to validate SSL 
        connections
        
        @rtype sslCACertDir: basestring
        @return sslCACertDirList: directory containing PEM encoded CA 
        certificates."""
        return self.__sslCACertDir
    
    def _setSSLCACertDir(self, value):
        """Set CA cert or certs to validate AC signatures, signatures
        of Attribute Authority SOAP responses and SSL connections where 
        AA SOAP service is run over SSL.
        
        @type sslCACertDir: basestring
        @param sslCACertDir: directory containing CA certificate files.
        """
        if isinstance(value, basestring):
            self.__sslCACertDir = os.path.expandvars(value)
        elif value is None:
            self.__sslCACertDir = value
        else:
            raise TypeError("Input CA Certificate directroy must be "
                            "a valid string or None type: %r" % type(value))      
        
    sslCACertDir = property(fget=_getSSLCACertDir,
                            fset=_setSSLCACertDir,
                            doc="Path to directory containing PEM encoded CA "
                                "Certificates used for verification of peer "
                                "certs in SSL connection.   Files in the "
                                "directory must be named with the form "
                                "<hash>.0 where <hash> can be obtained using "
                                "openssl x509 -in cert -hash -noout or using "
                                "the c_rehash OpenSSL script")
    
    def _getSslValidDNs(self):
        return self.__sslValidDNs

    def _setSslValidDNs(self, value):
        if isinstance(value, basestring):  
            pat = SSLContextProxy.VALID_DNS_PAT
            self.__sslValidDNs = [X500DN.fromString(dn) 
                                  for dn in pat.split(value)]
            
        elif isinstance(value, (tuple, list)):
            self.__sslValidDNs = [X500DN.fromString(dn) for dn in value]
        else:
            raise TypeError('Expecting list/tuple or basestring type for "%s" '
                            'attribute; got %r' %
                            (SSLContextProxy.SSL_VALID_DNS_OPTNAME, 
                             type(value)))
    
    sslValidDNs = property(_getSslValidDNs, 
                           _setSslValidDNs, 
                           doc="whitelist of acceptable certificate "
                               "Distinguished Names for peer certificates in "
                               "SSL requests")

    def _getSSLPriKeyFilePath(self):
        return self.__sslPriKeyFilePath
    
    def _setSSLPriKeyFilePath(self, filePath):
        "Set ssl private key file path property method"
        
        if isinstance(filePath, basestring):
            filePath = os.path.expandvars(filePath)

        elif filePath is not None:
            raise TypeError("Private key file path must be a valid "
                            "string or None type")
        
        self.__sslPriKeyFilePath = filePath
        
    sslPriKeyFilePath = property(fget=_getSSLPriKeyFilePath,
                                 fset=_setSSLPriKeyFilePath,
                                 doc="File path to SSL private key")
 
    def _setSSLPriKeyPwd(self, sslPriKeyPwd):
        "Set method for ssl private key file password"
        if not isinstance(sslPriKeyPwd, (type(None), basestring)):
            raise TypeError("Signing private key password must be None "
                            "or a valid string")
        
        # Explicitly convert to string as M2Crypto OpenSSL wrapper fails with
        # unicode type
        self.__sslPriKeyPwd = str(sslPriKeyPwd)

    def _getSSLPriKeyPwd(self):
        "Get property method for SSL private key"
        return self.__sslPriKeyPwd
        
    sslPriKeyPwd = property(fset=_setSSLPriKeyPwd,
                             fget=_getSSLPriKeyPwd,
                             doc="Password protecting SSL private key file")

    def __getstate__(self):
        '''Enable pickling for use with beaker.session'''
        _dict = {}
        for attrName in SSLContextProxy.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_SSLContextProxy" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
        
    def __setstate__(self, attrDict):
        '''Enable pickling for use with beaker.session'''
        for attr, val in attrDict.items():
            setattr(self, attr, val)


"""NDG Security SAML unit test package for ElementTree based SAML Attribute
Assertion implementation

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "25/07/08"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $'


"""SOAP client package - XML representation using ElementTree

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "27/07/09"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: etree.py 7131 2010-06-30 13:37:48Z pjkersha $'
import logging
log = logging.getLogger(__name__)
    
try: # python 2.5
    from xml.etree import cElementTree, ElementTree
except ImportError:
    # if you've installed it yourself it comes this way
    import cElementTree, ElementTree

# ElementTree helper functions
from ndg.soap.utils.etree import QName

from ndg.soap import (SOAPObject, SOAPEnvelopeBase, SOAPHeaderBase, 
                      SOAPBodyBase, SOAPFault)


class ETreeSOAPExtensions(object):  
    """Utility to enable addition of core ElementTree specific attributes and
    methods for ElementTree SOAP implementation
    """
    def __init__(self):
        self.__qname = None
        self.__elem = None

    def _getQname(self):
        return self.__qname

    def _setQname(self, value):
        if not isinstance(value, QName):
            raise TypeError('Expecting %r for "qname" attribute; got %r' %
                            (QName, type(value)))
        self.__qname = value

    def _getElem(self):
        return self.__elem

    def _setElem(self, value):
        if not ElementTree.iselement(value):
            raise TypeError('Expecting %r for "elem" attribute; got %r' %
                            (ElementTree.Element, type(value)))
        self.__elem = value
        
    qname = property(_getQname, _setQname, None, "Qualified name object")
    elem = property(_getElem, _setElem, None, "Root element")

    @staticmethod
    def _serialize(elem):
         """Serialise element tree into string"""
         return cElementTree.tostring(elem)
       
    @classmethod
    def _prettyPrint(cls, elem):
        """Basic pretty printing separating each element on to a new line"""
        xml = cls._serialize(elem)
        xml = ">\n".join(xml.split(">"))
        xml = "\n<".join(xml.split("<"))
        xml = '\n'.join(xml.split('\n\n'))
        return xml

    def _parse(self, source):
        """Read in the XML from source
        @type source: basestring/file
        @param source: file path to XML file or file object
        """
        tree = ElementTree.parse(source)
        elem = tree.getroot()
        
        return elem        


class SOAPHeader(SOAPHeaderBase, ETreeSOAPExtensions):
    """ElementTree implementation of SOAP Header object"""
    
    DEFAULT_ELEMENT_NAME = QName(SOAPHeaderBase.DEFAULT_ELEMENT_NS,
                               tag=SOAPHeaderBase.DEFAULT_ELEMENT_LOCAL_NAME,
                               prefix=SOAPHeaderBase.DEFAULT_ELEMENT_NS_PREFIX)
    
    def __init__(self):
        SOAPHeaderBase.__init__(self)
        ETreeSOAPExtensions.__init__(self)
        
        self.qname = QName(SOAPHeaderBase.DEFAULT_ELEMENT_NS, 
                           tag=SOAPHeaderBase.DEFAULT_ELEMENT_LOCAL_NAME, 
                           prefix=SOAPHeaderBase.DEFAULT_ELEMENT_NS_PREFIX)

    def create(self):
        """Create header ElementTree element"""
        
        self.elem = ElementTree.Element(str(self.qname))
        ElementTree._namespace_map[SOAPHeaderBase.DEFAULT_ELEMENT_NS
                                   ] = SOAPHeaderBase.DEFAULT_ELEMENT_NS_PREFIX
    
    def serialize(self):
        """Serialise element tree into string"""
        return ETreeSOAPExtensions._serialize(self.elem)
    
    def prettyPrint(self):
        """Basic pretty printing separating each element on to a new line"""
        return ETreeSOAPExtensions._prettyPrint(self.elem)


class SOAPBody(SOAPBodyBase, ETreeSOAPExtensions):
    """ElementTree based implementation for SOAP Body object"""
    
    DEFAULT_ELEMENT_NAME = QName(SOAPBodyBase.DEFAULT_ELEMENT_NS,
                                 tag=SOAPBodyBase.DEFAULT_ELEMENT_LOCAL_NAME,
                                 prefix=SOAPBodyBase.DEFAULT_ELEMENT_NS_PREFIX)
    
    def __init__(self):
        SOAPBodyBase.__init__(self)
        ETreeSOAPExtensions.__init__(self)
        
        self.qname = QName(SOAPBodyBase.DEFAULT_ELEMENT_NS, 
                           tag=SOAPBodyBase.DEFAULT_ELEMENT_LOCAL_NAME, 
                           prefix=SOAPBodyBase.DEFAULT_ELEMENT_NS_PREFIX)
        
    def create(self):
        """Create header ElementTree element"""
        self.elem = ElementTree.Element(str(self.qname))
    
    def serialize(self):
        """Serialise element tree into string"""
        return ETreeSOAPExtensions._serialize(self.elem)
    
    def prettyPrint(self):
        """Basic pretty printing separating each element on to a new line"""
        return ETreeSOAPExtensions._prettyPrint(self.elem)
    

class SOAPEnvelope(SOAPEnvelopeBase, ETreeSOAPExtensions):
    """ElementTree based SOAP implementation"""
    DEFAULT_ELEMENT_NAME = QName(SOAPEnvelopeBase.DEFAULT_ELEMENT_NS,
                             tag=SOAPEnvelopeBase.DEFAULT_ELEMENT_LOCAL_NAME,
                             prefix=SOAPEnvelopeBase.DEFAULT_ELEMENT_NS_PREFIX)

    def __init__(self):
        SOAPEnvelopeBase.__init__(self)
        ETreeSOAPExtensions.__init__(self)
        
        self.qname = QName(SOAPEnvelopeBase.DEFAULT_ELEMENT_NS, 
                           tag=SOAPEnvelopeBase.DEFAULT_ELEMENT_LOCAL_NAME, 
                           prefix=SOAPEnvelopeBase.DEFAULT_ELEMENT_NS_PREFIX)
        self.__header = SOAPHeader()
        self.__body = SOAPBody()

    def _getHeader(self):
        return self.__header

    def _setHeader(self, value):
        if not isinstance(value, SOAPHeader):
            raise TypeError('Expecting %r for "header" attribute; got %r' %
                            (SOAPHeader, type(value)))
        self.__header = value

    def _getBody(self):
        return self.__body

    def _setBody(self, value):
        if not isinstance(value, SOAPBody):
            raise TypeError('Expecting %r for "header" attribute; got %r' %
                            (SOAPBody, type(value)))
        self.__body = value

    header = property(_getHeader, _setHeader, None, "SOAP header object")
    body = property(_getBody, _setBody, None, "SOAP body object")

    def create(self):
        """Create SOAP Envelope with header and body"""
        
        self.elem = ElementTree.Element(str(self.qname))
            
        self.header.create()
        self.elem.append(self.header.elem)
        
        self.body.create()
        self.elem.append(self.body.elem)
    
    def serialize(self):
        """Serialise element tree into string"""
        return ETreeSOAPExtensions._serialize(self.elem)
    
    def prettyPrint(self):
        """Basic pretty printing separating each element onto a new line"""
        return ETreeSOAPExtensions._prettyPrint(self.elem)
    
    def parse(self, source):
        self.elem = ETreeSOAPExtensions._parse(self, source) 
        
        for elem in self.elem:
            localName = QName.getLocalPart(elem.tag)
            if localName == SOAPHeader.DEFAULT_ELEMENT_LOCAL_NAME:
                self.header.elem = elem
                
            elif localName == SOAPBody.DEFAULT_ELEMENT_LOCAL_NAME:
                self.body.elem = elem
                
            else:
                raise SOAPFault('Invalid child element in SOAP Envelope "%s" '
                                'for stream %r' % (localName, source))



"""SAML Common XML module

Implementation of SAML 2.0 for NDG Security

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "23/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: xml.py 7130 2010-06-30 13:33:07Z pjkersha $"
from ndg.saml.xml import XMLConstants    


class SAMLConstants(XMLConstants):
    '''XML related constants used in the SAML specifications.
    @cvar POST_METHOD: HTTP Request Method - POST.                           
    @type POST_METHOD: string                                                 
    @cvar GET_METHOD: HTTP Method - GET.                                     
    @type GET_METHOD: string                                                  
    @cvar SCHEMA_DIR: Directory, on the classpath, schemas are located in.   
    @type SCHEMA_DIR: string                                                  
    @cvar XML_SCHEMA_LOCATION: XML core schema system Id.                    
    @type XML_SCHEMA_LOCATION: string                                         
    @cvar XMLSIG_SCHEMA_LOCATION:  XML Signature schema Id.                  
    @type XMLSIG_SCHEMA_LOCATION: string                                      
    @cvar XMLENC_SCHEMA_LOCATION: XML Encryption schema Id.                  
    @type XMLENC_SCHEMA_LOCATION: string                                      
    @cvar SOAP11ENV_SCHEMA_LOCATION:  SOAP 1.1 schema Id.                    
    @type SOAP11ENV_SCHEMA_LOCATION: string                                   
    @cvar SOAP11ENV_NS:  SOAP 1.1 Envelope XML namespace.                    
    @type SOAP11ENV_NS: string                                                
    @cvar SOAP11ENV_PREFIX:  SOAP 1.1 Envelope QName prefix.                 
    @type SOAP11ENV_PREFIX: string                                            
    @cvar PAOS_NS:  Liberty PAOS XML Namespace.                              
    @type PAOS_NS: string                                                     
    @cvar PAOS_PREFIX:  Liberty PAOS QName prefix.                           
    @type PAOS_PREFIX: string                                                 
    @cvar SAML10_SCHEMA_LOCATION: SAML 1.0 Assertion schema system Id.       
    @type SAML10_SCHEMA_LOCATION: string                                      
    @cvar SAML11_SCHEMA_LOCATION: SAML 1.1 Assertion schema system Id.       
    @type SAML11_SCHEMA_LOCATION: string                                      
    @cvar SAML1_NS: SAML 1.X XML namespace.                                  
    @type SAML1_NS: string                                                    
    @cvar SAML10P_SCHEMA_LOCATION: SAML 1.0 Protocol schema system Id.       
    @type SAML10P_SCHEMA_LOCATION: string                                     
    @cvar SAML11P_SCHEMA_LOCATION: SAML 1.1 Protocol schema system Id.       
    @type SAML11P_SCHEMA_LOCATION: string                                     
    @cvar SAML10P_NS: SAML 1.X protocol XML namespace.                       
    @type SAML10P_NS: string                                                  
    @cvar SAML11P_NS: SupportEnumeration.                                    
    @type SAML11P_NS: string                                                  
    @cvar SAML1P_PREFIX: SAML 1.X Protocol QName prefix.                     
    @type SAML1P_PREFIX: string                                               
    @cvar SAML1_PREFIX: SAML 1.X Assertion QName prefix.                     
    @type SAML1_PREFIX: string                                                
    @cvar SAML1MD_NS: SAML 1 Metadata extension XML namespace.               
    @type SAML1MD_NS: string                                                  
    @cvar SAML1MD_SCHEMA_LOCATION: SAML 1 Metadata extension schema system Id.
    @type SAML1MD_SCHEMA_LOCATION: string                                      
    @cvar SAML1MD_PREFIX: SAML 1 Metadata extension namespace prefix.         
    @type SAML1MD_PREFIX: string                                               
    @cvar SAML1_ARTIFACT_BINDING_URI: URI for SAML 1 Artifact binding.        
    @type SAML1_ARTIFACT_BINDING_URI: string                                   
    @cvar SAML1_POST_BINDING_URI: URI for SAML 1 POST binding.                
    @type SAML1_POST_BINDING_URI: string                                       
    @cvar SAML1_SOAP11_BINDING_URI: URI for SAML 1 SOAP 1.1 binding.          
    @type SAML1_SOAP11_BINDING_URI: string                                     
    @cvar SAML20_SCHEMA_LOCATION: SAML 2.0 Assertion schema Id.               
    @type SAML20_SCHEMA_LOCATION: string                                       
    @cvar SAML20_NS: SAML 2.0 Assertion XML Namespace.                        
    @type SAML20_NS: string                                                    
    @cvar SAML20_PREFIX: SAML 2.0 Assertion QName prefix.                     
    @type SAML20_PREFIX: string                                                
    @cvar SAML20P_SCHEMA_LOCATION: SAML 2.0 Protocol schema Id.               
    @type SAML20P_SCHEMA_LOCATION: string                                      
    @cvar SAML20P_NS: SAML 2.0 Protocol XML Namespace.                        
    @type SAML20P_NS: string                                                   
    @cvar SAML20P_PREFIX: SAML 2.0 Protocol QName prefix.                     
    @type SAML20P_PREFIX: string                                               
    @cvar SAML20PTHRPTY_SCHEMA_LOCATION: SAML 2.0 Protocol Third-party extension schema Id.
    @type SAML20PTHRPTY_SCHEMA_LOCATION: string                                             
    @cvar SAML20PTHRPTY_NS: SAML 2.0 Protocol XML Namespace.                               
    @type SAML20PTHRPTY_NS: string                                                          
    @cvar SAML20PTHRPTY_PREFIX: SAML 2.0 Protocol QName prefix.                            
    @type SAML20PTHRPTY_PREFIX: string                                                      
    @cvar SAML20MD_SCHEMA_LOCATION: SAML 2.0 Metadata schema Id.                           
    @type SAML20MD_SCHEMA_LOCATION: string                                                  
    @cvar SAML20MD_NS: SAML 2.0 Metadata XML Namespace.                                    
    @type SAML20MD_NS: string                                                               
    @cvar SAML20MDQUERY_NS: SAML 2.0 Standalone Query Metadata extension XML namespace.    
    @type SAML20MDQUERY_NS: string                                                          
    @cvar SAML20MDQUERY_SCHEMA_LOCATION: SAML 2.0 Standalone Query Metadata extension schema system Id.
    @type SAML20MDQUERY_SCHEMA_LOCATION: string                                                         
    @cvar SAML20MDQUERY_PREFIX: SAML 2.0 Standalone Query Metadata extension prefix.                   
    @type SAML20MDQUERY_PREFIX: string                                                                  
    @cvar SAML20MD_PREFIX: SAML 2.0 Metadata QName prefix.
    @type SAML20MD_PREFIX: string
    @cvar SAML20AC_SCHEMA_LOCATION: SAML 2.0 Authentication Context schema Id.
    @type SAML20AC_SCHEMA_LOCATION: string
    @cvar SAML20AC_NS: SAML 2.0 Authentication Context XML Namespace.
    @type SAML20AC_NS: string
    @cvar SAML20AC_PREFIX: SAML 2.0 Authentication Context QName prefix.
    @type SAML20AC_PREFIX: string
    @cvar SAML20ECP_SCHEMA_LOCATION: SAML 2.0 Enhanced Client/Proxy SSO Profile schema Id.
    @type SAML20ECP_SCHEMA_LOCATION: string
    @cvar SAML20ECP_NS: SAML 2.0 Enhanced Client/Proxy SSO Profile XML Namespace.
    @type SAML20ECP_NS: string
    @cvar SAML20ECP_PREFIX: SAML 2.0 Enhanced Client/Proxy SSO Profile QName prefix.
    @type SAML20ECP_PREFIX: string
    @cvar SAML20DCE_SCHEMA_LOCATION: SAML 2.0 DCE PAC Attribute Profile schema Id.
    @type SAML20DCE_SCHEMA_LOCATION: string
    @cvar SAML20DCE_NS: SAML 2.0 DCE PAC Attribute Profile XML Namespace.
    @type SAML20DCE_NS: string
    @cvar SAML20DCE_PREFIX: SAML 2.0 DCE PAC Attribute Profile QName prefix.
    @type SAML20DCE_PREFIX: string
    @cvar SAML20X500_SCHEMA_LOCATION: SAML 2.0 X.500 Attribute Profile schema Id.
    @type SAML20X500_SCHEMA_LOCATION: string
    @cvar SAML20X500_NS: SAML 2.0 X.500 Attribute Profile XML Namespace.
    @type SAML20X500_NS: string
    @cvar SAML20X500_PREFIX: SAML 2.0 X.500 Attribute Profile QName prefix.
    @type SAML20X500_PREFIX: string
    @cvar SAML20XACML_SCHEMA_LOCATION: SAML 2.0 XACML Attribute Profile schema Id.
    @type SAML20XACML_SCHEMA_LOCATION: string
    @cvar SAML20XACML_NS: SAML 2.0 XACML Attribute Profile XML Namespace.
    @type SAML20XACML_NS: string
    @cvar SAML20XACML_PREFIX: SAML 2.0 XACML Attribute Profile QName prefix.
    @type SAML20XACML_PREFIX: string
    @cvar SAML2_ARTIFACT_BINDING_URI: URI for SAML 2 Artifact binding.
    @type SAML2_ARTIFACT_BINDING_URI: string
    @cvar SAML2_POST_BINDING_URI: URI for SAML 2 POST binding.
    @type SAML2_POST_BINDING_URI: string
    @cvar SAML2_POST_SIMPLE_SIGN_BINDING_URI: URI for SAML 2 POST-SimpleSign binding.
    @type SAML2_POST_SIMPLE_SIGN_BINDING_URI: string
    @cvar SAML2_REDIRECT_BINDING_URI: URI for SAML 2 HTTP redirect binding.
    @type SAML2_REDIRECT_BINDING_URI: string
    @cvar SAML2_SOAP11_BINDING_URI: URI for SAML 2 SOAP binding.
    @type SAML2_SOAP11_BINDING_URI: string
    '''
    
    # HTTP Constants
    
    # HTTP Request Method - POST.
    POST_METHOD = "POST"
    
    # HTTP Method - GET.
    GET_METHOD = "GET"
    
    # OpenSAML 2
    
    # Directory, on the classpath, schemas are located in.
    SCHEMA_DIR = "/schema/"
    
    #    Core XML
    
    # XML core schema system Id.
    XML_SCHEMA_LOCATION = SCHEMA_DIR + "xml.xsd"
    
    #  XML Signature schema Id.
    XMLSIG_SCHEMA_LOCATION = SCHEMA_DIR + "xmldsig-core-schema.xsd"
    
    # XML Encryption schema Id.
    XMLENC_SCHEMA_LOCATION = SCHEMA_DIR + "xenc-schema.xsd"

    
    #    SOAP
    
    #  SOAP 1.1 schema Id.
    SOAP11ENV_SCHEMA_LOCATION = SCHEMA_DIR + SCHEMA_DIR + "soap-envelope.xsd"
    
    #  SOAP 1.1 Envelope XML namespace.
    SOAP11ENV_NS = "http://schemas.xmlsoap.org/soap/envelope/"
    
    #  SOAP 1.1 Envelope QName prefix.
    SOAP11ENV_PREFIX = "SOAP-ENV"
    
    #  Liberty PAOS XML Namespace.
    PAOS_NS = "urn:liberty:paos:2003-08"
    
    #  Liberty PAOS QName prefix.
    PAOS_PREFIX = "paos"
    
    #    SAML 1.X
    
    # SAML 1.0 Assertion schema system Id.
    SAML10_SCHEMA_LOCATION = SCHEMA_DIR + "cs-sstc-schema-assertion-01.xsd"
    
    # SAML 1.1 Assertion schema system Id.
    SAML11_SCHEMA_LOCATION = SCHEMA_DIR + "cs-sstc-schema-assertion-1.1.xsd"
    
    # SAML 1.X XML namespace.
    SAML1_NS = "urn:oasis:names:tc:SAML:1.0:assertion"
    
    # SAML 1.0 Protocol schema system Id.
    SAML10P_SCHEMA_LOCATION = SCHEMA_DIR + "cs-sstc-schema-protocol-01.xsd"
    
    # SAML 1.1 Protocol schema system Id.
    SAML11P_SCHEMA_LOCATION = SCHEMA_DIR + "cs-sstc-schema-protocol-1.1.xsd"

    # SAML 1.X protocol XML namespace.
    SAML10P_NS = "urn:oasis:names:tc:SAML:1.0:protocol"
    
    # SAML 1.1 protocol XML namespace, used only in SAML 2 metadata protocol
    # SupportEnumeration.
    SAML11P_NS = "urn:oasis:names:tc:SAML:1.1:protocol"
    
    # SAML 1.X Protocol QName prefix.
    SAML1P_PREFIX = "samlp"

    # SAML 1.X Assertion QName prefix.
    SAML1_PREFIX = "saml"
    
    # SAML 1 Metadata extension XML namespace.
    SAML1MD_NS = "urn:oasis:names:tc:SAML:profiles:v1metadata"
    
    # SAML 1 Metadata extension schema system Id.
    SAML1MD_SCHEMA_LOCATION = SCHEMA_DIR + "sstc-saml1x-metadata.xsd"
    
    # SAML 1 Metadata extension namespace prefix.
    SAML1MD_PREFIX = "saml1md"
    
    # URI for SAML 1 Artifact binding.
    SAML1_ARTIFACT_BINDING_URI = \
        "urn:oasis:names:tc:SAML:1.0:profiles:artifact-01"
    
    # URI for SAML 1 POST binding.
    SAML1_POST_BINDING_URI = \
        "urn:oasis:names:tc:SAML:1.0:profiles:browser-post"
    
    # URI for SAML 1 SOAP 1.1 binding.
    SAML1_SOAP11_BINDING_URI = \
        "urn:oasis:names:tc:SAML:1.0:bindings:SOAP-binding"
    
    #    SAML 2.0
    
    # SAML 2.0 Assertion schema Id.
    SAML20_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-assertion-2.0.xsd"
    
    # SAML 2.0 Assertion XML Namespace.
    SAML20_NS = "urn:oasis:names:tc:SAML:2.0:assertion"
    
    # SAML 2.0 Assertion QName prefix.
    SAML20_PREFIX ="saml"
    
    # SAML 2.0 Protocol schema Id.
    SAML20P_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-protocol-2.0.xsd"
    
    # SAML 2.0 Protocol XML Namespace.
    SAML20P_NS = "urn:oasis:names:tc:SAML:2.0:protocol"
    
    # SAML 2.0 Protocol QName prefix.
    SAML20P_PREFIX ="samlp"
    
    # SAML 2.0 Protocol Third-party extension schema Id.
    SAML20PTHRPTY_SCHEMA_LOCATION = SCHEMA_DIR + \
                                    "sstc-saml-protocol-ext-thirdparty.xsd"
    
    # SAML 2.0 Protocol XML Namespace.
    SAML20PTHRPTY_NS = "urn:oasis:names:tc:SAML:protocol:ext:third-party"
    
    # SAML 2.0 Protocol QName prefix.
    SAML20PTHRPTY_PREFIX ="thrpty"
    
    # SAML 2.0 Metadata schema Id.
    SAML20MD_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-metadata-2.0.xsd"
    
    # SAML 2.0 Metadata XML Namespace.
    SAML20MD_NS ="urn:oasis:names:tc:SAML:2.0:metadata"
    
    # SAML 2.0 Standalone Query Metadata extension XML namespace.
    SAML20MDQUERY_NS = "urn:oasis:names:tc:SAML:metadata:ext:query"
    
    # SAML 2.0 Standalone Query Metadata extension schema system Id.
    SAML20MDQUERY_SCHEMA_LOCATION = SCHEMA_DIR + \
                                    "sstc-saml-metadata-ext-query.xsd"
    
    # SAML 2.0 Standalone Query Metadata extension prefix.
    SAML20MDQUERY_PREFIX = "query"
    
    # SAML 2.0 Metadata QName prefix.
    SAML20MD_PREFIX = "md"
    
    # SAML 2.0 Authentication Context schema Id.
    SAML20AC_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-authn-context-2.0.xsd"
    
    # SAML 2.0 Authentication Context XML Namespace.
    SAML20AC_NS ="urn:oasis:names:tc:SAML:2.0:ac"
    
    # SAML 2.0 Authentication Context QName prefix.
    SAML20AC_PREFIX = "ac"
    
    # SAML 2.0 Enhanced Client/Proxy SSO Profile schema Id.
    SAML20ECP_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-ecp-2.0.xsd"
    
    # SAML 2.0 Enhanced Client/Proxy SSO Profile XML Namespace.
    SAML20ECP_NS = "urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp"
    
    # SAML 2.0 Enhanced Client/Proxy SSO Profile QName prefix.
    SAML20ECP_PREFIX = "ecp"
    
    # SAML 2.0 DCE PAC Attribute Profile schema Id.
    SAML20DCE_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-dce-2.0.xsd"
    
    # SAML 2.0 DCE PAC Attribute Profile XML Namespace.
    SAML20DCE_NS = "urn:oasis:names:tc:SAML:2.0:profiles:attribute:DCE"
    
    # SAML 2.0 DCE PAC Attribute Profile QName prefix.
    SAML20DCE_PREFIX = "DCE"
    
    # SAML 2.0 X.500 Attribute Profile schema Id.
    SAML20X500_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-x500-2.0.xsd"
    
    # SAML 2.0 X.500 Attribute Profile XML Namespace.
    SAML20X500_NS = "urn:oasis:names:tc:SAML:2.0:profiles:attribute:X500"
    
    # SAML 2.0 X.500 Attribute Profile QName prefix.
    SAML20X500_PREFIX = "x500"
    
    # SAML 2.0 XACML Attribute Profile schema Id.
    SAML20XACML_SCHEMA_LOCATION = SCHEMA_DIR + "saml-schema-xacml-2.0.xsd"
    
    # SAML 2.0 XACML Attribute Profile XML Namespace.
    SAML20XACML_NS = "urn:oasis:names:tc:SAML:2.0:profiles:attribute:XACML"
    
    # SAML 2.0 XACML Attribute Profile QName prefix.
    SAML20XACML_PREFIX = "xacmlprof"
    
    # URI for SAML 2 Artifact binding.
    SAML2_ARTIFACT_BINDING_URI = \
                        "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact"
    
    # URI for SAML 2 POST binding.
    SAML2_POST_BINDING_URI = "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
    
    # URI for SAML 2 POST-SimpleSign binding.
    SAML2_POST_SIMPLE_SIGN_BINDING_URI = \
                "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST-SimpleSign"
    
    # URI for SAML 2 HTTP redirect binding.
    SAML2_REDIRECT_BINDING_URI = \
                "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    
    # URI for SAML 2 SOAP binding.
    SAML2_SOAP11_BINDING_URI = "urn:oasis:names:tc:SAML:2.0:bindings:SOAP"
  
    
class QName(object):
    """XML Qualified Name
    @ivar __namespaceURI: the namespace the element is in
    @type __namespaceURI: basestring
    @ivar __localPart: the local name of the XML element 
    @type __localPart: basestring
    @ivar __prefix: the prefix for the given namespace
    @type __prefix: basestring
    """ 

    def __init__(self, namespaceURI, localPart, prefix):
        '''
        @param namespaceURI: the namespace the element is in
        @type namespaceURI: basestring
        @param localPart: the local name of the XML element 
        @type localPart: basestring
        @param prefix: the prefix for the given namespace
        @type prefix: basestring
        '''
        self.namespaceURI = namespaceURI
        self.localPart = localPart
        self.prefix = prefix
    
    def _getPrefix(self):
        """Get prefix
        @return: prefix
        @rtype: string
        """
        return self.__prefix

    def _setPrefix(self, value):
        """Set prefix
        @param value: prefix
        @type value: string
        @raise TypeError: invalid input value type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expected string type for "prefix"; got %r' %
                            type(value))
        self.__prefix = value
    
    prefix = property(_getPrefix, _setPrefix, None, "Namespace Prefix")

    def _getLocalPart(self):
        """Get local part
        @return: local part
        @rtype: string
        """
        return self.__localPart
    
    def _setLocalPart(self, value):
        """Set local part
        @param value: local part
        @type value: string
        @raise TypeError: invalid input value type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expected string type for "localPart"; got %r' %
                            type(value))
        self.__localPart = value
        
    localPart = property(_getLocalPart, _setLocalPart, None, "LocalPart")

    def _getNamespaceURI(self):
        """Get namespace URI
        @return: namespace URI
        @rtype: string
        """
        return self.__namespaceURI

    def _setNamespaceURI(self, value):
        """Set namespace URI
        @param value: namespace URI
        @type value: string
        @raise TypeError: invalid input value type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expected string type for "namespaceURI"; got %r' %
                            type(value))
        self.__namespaceURI = value
  
    namespaceURI = property(_getNamespaceURI, _setNamespaceURI, None, 
                            "Namespace URI")

    def __eq__(self, qname):
        """Enable equality check for QName
        @type qname: saml.common.xml.QName
        @param qname: Qualified Name to compare with self 
        @return: True if input and this object match
        @rtype: bool
        @return: True if input and this object match
        @rtype: bool
        """
        if not isinstance(qname, QName):
            raise TypeError('Expecting %r; got %r' % (QName, type(qname)))
                            
        return (self.prefix, self.namespaceURI, self.localPart) == \
               (qname.prefix, qname.namespaceURI, qname.localPart)

    def __ne__(self, qname):
        """Enable equality check for QName
        @type qname: saml.common.xml.QName
        @param qname: Qualified Name to compare with self 
        @return: True if input and this object don't match
        @rtype: bool
        """
        return not self.__eq__(qname)



#!/usr/bin/env python
"""Unit tests for WSGI SAML 2.0 SOAP Attribute Query Interface

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "21/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: test_attributeservice.py 7165 2010-07-13 08:32:36Z pjkersha $'
import unittest
from uuid import uuid4
from datetime import datetime, timedelta
from cStringIO import StringIO

from ndg.soap.etree import SOAPEnvelope

from ndg.saml.saml2.core import (Assertion, Attribute, AttributeStatement, 
                                 SAMLVersion, Subject, NameID, Issuer, 
                                 AttributeQuery, XSStringAttributeValue, 
                                 Conditions, Status, StatusCode)
from ndg.saml.xml import XMLConstants
from ndg.saml.xml.etree import AttributeQueryElementTree, ResponseElementTree
from ndg.saml.test.binding.soap import WithPasteFixtureBaseTestCase


class TestAttributeServiceMiddleware(object):
    """Test Attribute Service interface stub"""
    QUERY_INTERFACE_KEYNAME_OPTNAME = 'queryInterfaceKeyName'
    ISSUER_DN = '/O=Test/OU=Attribute Service/CN=Service Stub'
    
    FIRSTNAME_ATTRNAME = "urn:ndg:saml:firstname"
    LASTNAME_ATTRNAME = "urn:ndg:saml:lastname"
    EMAILADDRESS_ATTRNAME = "urn:ndg:saml:emailaddress"
    
    VALID_QUERY_ISSUERS = (
        "/O=Site A/CN=Authorisation Service",
        "/O=Site B/CN=Authorisation Service"
    )
    VALID_SUBJECTS = ("https://openid.localhost/philip.kershaw", )
    VALID_ATTR_NAME_URIS = (
        FIRSTNAME_ATTRNAME, LASTNAME_ATTRNAME, EMAILADDRESS_ATTRNAME
    )
    
    def __init__(self, app, global_conf, **app_conf):
        self.queryInterfaceKeyName = app_conf[
            self.__class__.QUERY_INTERFACE_KEYNAME_OPTNAME]
        self._app = app

        self.firstName = "Philip"
        self.lastName = "Kershaw"
        self.emailAddress = "pkershaw@somewhere.ac.uk"
    
    def __call__(self, environ, start_response):
        environ[self.queryInterfaceKeyName] = self.attributeQueryFactory()
        return self._app(environ, start_response)
    
    def attributeQueryFactory(self):
        """Makes the attribute query method"""
        
        def attributeQuery(query, response):
            """Attribute Query interface called by the next middleware in the 
            stack the SAML SOAP Query interface middleware instance
            (ndg.saml.saml2.binding.soap.server.wsgi.queryinterface.SOAPQueryInterfaceMiddleware)
            """
            response.issueInstant = datetime.utcnow()
            response.id = str(uuid4())
            response.issuer = Issuer()
            
            # SAML 2.0 spec says format must be omitted
            #response.issuer.format = Issuer.X509_SUBJECT
            response.issuer.value = \
                            "/O=NDG/OU=BADC/CN=attributeauthority.badc.rl.ac.uk"
            
            response.inResponseTo = query.id

            if query.issuer.value not in self.__class__.VALID_QUERY_ISSUERS:
                response.status.statusCode.value = \
                                            StatusCode.REQUEST_DENIED_URI   
                return response    

            if query.subject.nameID.value not in self.__class__.VALID_SUBJECTS:
                response.status.statusCode.value = \
                                            StatusCode.UNKNOWN_PRINCIPAL_URI   
                return response    
                
            assertion = Assertion()
            
            assertion.version = SAMLVersion(SAMLVersion.VERSION_20)
            assertion.id = str(uuid4())
            assertion.issueInstant = response.issueInstant
            
            assertion.conditions = Conditions()
            assertion.conditions.notBefore = assertion.issueInstant
            assertion.conditions.notOnOrAfter = \
                assertion.conditions.notBefore + timedelta(seconds=60*60*8)
            
            assertion.subject = Subject()  
            assertion.subject.nameID = NameID()
            assertion.subject.nameID.format = query.subject.nameID.format
            assertion.subject.nameID.value = query.subject.nameID.value
    
            assertion.attributeStatements.append(AttributeStatement())                
            
            for attribute in query.attributes:
                if attribute.name == self.__class__.FIRSTNAME_ATTRNAME:
                    # special case handling for 'FirstName' attribute
                    fnAttribute = Attribute()
                    fnAttribute.name = attribute.name
                    fnAttribute.nameFormat = attribute.nameFormat
                    fnAttribute.friendlyName = attribute.friendlyName
        
                    firstName = XSStringAttributeValue()
                    firstName.value = self.firstName
                    fnAttribute.attributeValues.append(firstName)
        
                    assertion.attributeStatements[0].attributes.append(
                                                                    fnAttribute)
                
                elif attribute.name == self.__class__.LASTNAME_ATTRNAME:
                    lnAttribute = Attribute()
                    lnAttribute.name = attribute.name
                    lnAttribute.nameFormat = attribute.nameFormat
                    lnAttribute.friendlyName = attribute.friendlyName
        
                    lastName = XSStringAttributeValue()
                    lastName.value = self.lastName
                    lnAttribute.attributeValues.append(lastName)
        
                    assertion.attributeStatements[0].attributes.append(
                                                                    lnAttribute)
                   
                elif (attribute.name == self.__class__.EMAILADDRESS_ATTRNAME and
                      query.issuer.value == 
                                        self.__class__.VALID_QUERY_ISSUERS[0]):
                    emailAddressAttribute = Attribute()
                    emailAddressAttribute.name = attribute.name
                    emailAddressAttribute.nameFormat = attribute.nameFormat
                    emailAddressAttribute.friendlyName = attribute.friendlyName
        
                    emailAddress = XSStringAttributeValue()
                    emailAddress.value = self.emailAddress
                    emailAddressAttribute.attributeValues.append(emailAddress)
        
                    assertion.attributeStatements[0].attributes.append(
                                                        emailAddressAttribute)
                else:
                    response.status.statusCode.value = \
                                        StatusCode.INVALID_ATTR_NAME_VALUE_URI
                    return response
                                    
            
            response.assertions.append(assertion)
            response.status.statusCode.value = StatusCode.SUCCESS_URI        
    
            return response
        
        return attributeQuery


class SOAPAttributeInterfaceMiddlewareTestCase(WithPasteFixtureBaseTestCase):
    """Test SAML Attribute Query over SOAP Binding querying a test attribute
    server served using Paste Paster over HTTPS""" 
    CONFIG_FILENAME = 'attribute-interface.ini'
    SERVICE_URI = '/attributeauthority'
    
    def _createAttributeQuery(self, 
                        issuer="/O=Site A/CN=Authorisation Service",
                        subject="https://openid.localhost/philip.kershaw"):
        attributeQuery = AttributeQuery()
        attributeQuery.version = SAMLVersion(SAMLVersion.VERSION_20)
        attributeQuery.id = str(uuid4())
        attributeQuery.issueInstant = datetime.utcnow()
        
        attributeQuery.issuer = Issuer()
        attributeQuery.issuer.format = Issuer.X509_SUBJECT
        attributeQuery.issuer.value = issuer
                        
        attributeQuery.subject = Subject()  
        attributeQuery.subject.nameID = NameID()
        attributeQuery.subject.nameID.format = "urn:ndg:saml:test:openid"
        attributeQuery.subject.nameID.value = subject
                                    
        
        # special case handling for 'FirstName' attribute
        fnAttribute = Attribute()
        fnAttribute.name = TestAttributeServiceMiddleware.FIRSTNAME_ATTRNAME
        fnAttribute.nameFormat = "http://www.w3.org/2001/XMLSchema#string"
        fnAttribute.friendlyName = "FirstName"

        attributeQuery.attributes.append(fnAttribute)
    
        # special case handling for 'LastName' attribute
        lnAttribute = Attribute()
        lnAttribute.name = TestAttributeServiceMiddleware.LASTNAME_ATTRNAME
        lnAttribute.nameFormat = "http://www.w3.org/2001/XMLSchema#string"
        lnAttribute.friendlyName = "LastName"

        attributeQuery.attributes.append(lnAttribute)
    
        # special case handling for 'LastName' attribute
        emailAddressAttribute = Attribute()
        emailAddressAttribute.name = \
                            TestAttributeServiceMiddleware.EMAILADDRESS_ATTRNAME
        emailAddressAttribute.nameFormat = XMLConstants.XSD_NS+"#"+\
                                    XSStringAttributeValue.TYPE_LOCAL_NAME
        emailAddressAttribute.friendlyName = "emailAddress"

        attributeQuery.attributes.append(emailAddressAttribute)  

        return attributeQuery
    
    def _makeRequest(self, attributeQuery=None, **kw):
        """Convenience method to construct queries for tests"""
        
        if attributeQuery is None:
            attributeQuery = self._createAttributeQuery(**kw)
            
        elem = AttributeQueryElementTree.toXML(attributeQuery)
        soapRequest = SOAPEnvelope()
        soapRequest.create()
        soapRequest.body.elem.append(elem)
        
        request = soapRequest.serialize()
        
        return request
    
    def _getSAMLResponse(self, responseBody):
        """Deserialise response string into ElementTree element"""
        soapResponse = SOAPEnvelope()
        
        responseStream = StringIO()
        responseStream.write(responseBody)
        responseStream.seek(0)
        
        soapResponse.parse(responseStream)
        
        print("Parsed response ...")
        print(soapResponse.serialize())
#        print(prettyPrint(soapResponse.elem))
        
        response = ResponseElementTree.fromXML(soapResponse.body.elem[0])
        
        return response
    
    def test01ValidQuery(self):
        attributeQuery = self._createAttributeQuery()
        request = self._makeRequest(attributeQuery=attributeQuery)
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        response = self.app.post(self.__class__.SERVICE_URI, 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        print("Response status=%d" % response.status)
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.SUCCESS_URI)
        self.assert_(samlResponse.inResponseTo == attributeQuery.id)
        self.assert_(samlResponse.assertions[0].subject.nameID.value == \
                     attributeQuery.subject.nameID.value)

    def test02AttributeReleaseDenied(self):
        request = self._makeRequest(issuer="/O=Site B/CN=Authorisation Service")
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        
        response = self.app.post(self.__class__.SERVICE_URI, 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.INVALID_ATTR_NAME_VALUE_URI)

    def test03InvalidAttributesRequested(self):
        attributeQuery = self._createAttributeQuery()
        
        # Add an unsupported Attribute name
        attribute = Attribute()
        attribute.name = "urn:my:attribute"
        attribute.nameFormat = XMLConstants.XSD_NS+"#"+\
                                    XSStringAttributeValue.TYPE_LOCAL_NAME
        attribute.friendlyName = "myAttribute"
        attributeQuery.attributes.append(attribute)     
        
        request = self._makeRequest(attributeQuery=attributeQuery)
           
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
       
        response = self.app.post(self.__class__.SERVICE_URI, 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.INVALID_ATTR_NAME_VALUE_URI)
        
    def test04InvalidQueryIssuer(self):
        request = self._makeRequest(issuer="/CN=My Attribute Query Issuer")
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
       
        response = self.app.post(self.__class__.SERVICE_URI, 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.REQUEST_DENIED_URI)

    def test05UnknownPrincipal(self):
        request = self._makeRequest(subject="Joe.Bloggs")
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        
        response = self.app.post(self.__class__.SERVICE_URI, 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.UNKNOWN_PRINCIPAL_URI)

 
if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
"""Unit tests for WSGI SAML 2.0 SOAP Authorisation Decision Query Interface

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "15/02/2010"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: test_authzservice.py 7154 2010-07-01 15:59:59Z pjkersha $'
import unittest
from uuid import uuid4
from datetime import datetime, timedelta
from cStringIO import StringIO

from ndg.soap.etree import SOAPEnvelope

from ndg.saml.saml2.core import (SAMLVersion, Subject, NameID, Issuer, 
                                 AuthzDecisionQuery, AuthzDecisionStatement, 
                                 Status, StatusCode, StatusMessage, 
                                 DecisionType, Action, Conditions, Assertion)
from ndg.saml.xml.etree import (AuthzDecisionQueryElementTree, 
                                ResponseElementTree)
from ndg.saml.test.binding.soap import WithPasteFixtureBaseTestCase


class TestAuthorisationServiceMiddleware(object):
    """Test Authorisation Service interface stub"""
    QUERY_INTERFACE_KEYNAME_OPTNAME = 'queryInterfaceKeyName'
    RESOURCE_URI = 'http://localhost/dap/data/'
    ISSUER_DN = '/O=Test/OU=Authorisation/CN=Service Stub'
    
    def __init__(self, app, global_conf, **app_conf):
        self.queryInterfaceKeyName = app_conf[
            TestAuthorisationServiceMiddleware.QUERY_INTERFACE_KEYNAME_OPTNAME]
        self._app = app
    
    def __call__(self, environ, start_response):
        environ[self.queryInterfaceKeyName] = self.authzDecisionQueryFactory()
        return self._app(environ, start_response)
    
    def authzDecisionQueryFactory(self):
        """Makes the authorisation decision"""
        
        def authzDecisionQuery(query, response):
            """Authorisation Decision Query interface called by the next 
            middleware in the stack the SAML SOAP Query interface middleware 
            instance
            (ndg.saml.saml2.binding.soap.server.wsgi.queryinterface.SOAPQueryInterfaceMiddleware)
            """
            now = datetime.utcnow()
            response.issueInstant = now
            
            # Make up a request ID that this response is responding to
            response.inResponseTo = query.id
            response.id = str(uuid4())
            response.version = SAMLVersion(SAMLVersion.VERSION_20)
            
            response.status = Status()
            response.status.statusCode = StatusCode()
            response.status.statusCode.value = StatusCode.SUCCESS_URI
            response.status.statusMessage = StatusMessage()        
            response.status.statusMessage.value = \
                                                "Response created successfully"
               
            assertion = Assertion()
            assertion.version = SAMLVersion(SAMLVersion.VERSION_20)
            assertion.id = str(uuid4())
            assertion.issueInstant = now
            
            authzDecisionStatement = AuthzDecisionStatement()
            
            # Make some simple logic to simulate a full access policy
            if query.resource == self.__class__.RESOURCE_URI:
                if query.actions[0].value == Action.HTTP_GET_ACTION:
                    authzDecisionStatement.decision = DecisionType.PERMIT
                else:
                    authzDecisionStatement.decision = DecisionType.DENY
            else:
                authzDecisionStatement.decision = DecisionType.INDETERMINATE
                
            authzDecisionStatement.resource = query.resource
                
            authzDecisionStatement.actions.append(Action())
            authzDecisionStatement.actions[-1].namespace = Action.GHPP_NS_URI
            authzDecisionStatement.actions[-1].value = Action.HTTP_GET_ACTION
            assertion.authzDecisionStatements.append(authzDecisionStatement)
            
            # Add a conditions statement for a validity of 8 hours
            assertion.conditions = Conditions()
            assertion.conditions.notBefore = now
            assertion.conditions.notOnOrAfter = now + timedelta(seconds=60*60*8)
                   
            assertion.subject = Subject()  
            assertion.subject.nameID = NameID()
            assertion.subject.nameID.format = query.subject.nameID.format
            assertion.subject.nameID.value = query.subject.nameID.value
                
            assertion.issuer = Issuer()
            assertion.issuer.format = Issuer.X509_SUBJECT
            assertion.issuer.value = \
                                    TestAuthorisationServiceMiddleware.ISSUER_DN
    
            response.assertions.append(assertion)
            return response
        
        return authzDecisionQuery
    
    
class SOAPAuthzDecisionInterfaceMiddlewareTestCase(
                                        WithPasteFixtureBaseTestCase):
    CONFIG_FILENAME = 'authz-decision-interface.ini'
    RESOURCE_URI = TestAuthorisationServiceMiddleware.RESOURCE_URI
    
    def _createAuthzDecisionQuery(self, 
                            issuer="/O=Site A/CN=PEP",
                            subject="https://openid.localhost/philip.kershaw",
                            resource=RESOURCE_URI,
                            action=Action.HTTP_GET_ACTION,
                            actionNs=Action.GHPP_NS_URI):
        query = AuthzDecisionQuery()
        query.version = SAMLVersion(SAMLVersion.VERSION_20)
        query.id = str(uuid4())
        query.issueInstant = datetime.utcnow()
        
        query.issuer = Issuer()
        query.issuer.format = Issuer.X509_SUBJECT
        query.issuer.value = issuer
                        
        query.subject = Subject()  
        query.subject.nameID = NameID()
        query.subject.nameID.format = "urn:ndg:saml:test:openid"
        query.subject.nameID.value = subject
   
        query.resource = resource
                 
        query.actions.append(Action())
        query.actions[0].namespace = actionNs
        query.actions[0].value = action    

        return query
    
    def _makeRequest(self, query=None, **kw):
        """Convenience method to construct queries for tests"""
        
        if query is None:
            query = self._createAuthzDecisionQuery(**kw)
            
        elem = AuthzDecisionQueryElementTree.toXML(query)
        soapRequest = SOAPEnvelope()
        soapRequest.create()
        soapRequest.body.elem.append(elem)
        
        request = soapRequest.serialize()
        
        return request
    
    def _getSAMLResponse(self, responseBody):
        """Deserialise response string into ElementTree element"""
        soapResponse = SOAPEnvelope()
        
        responseStream = StringIO()
        responseStream.write(responseBody)
        responseStream.seek(0)
        
        soapResponse.parse(responseStream)
        
        print("Parsed response ...")
        print(soapResponse.serialize())
#        print(prettyPrint(soapResponse.elem))
        
        response = ResponseElementTree.fromXML(soapResponse.body.elem[0])
        
        return response
    
    def test01AccessGranted(self):
        query = self._createAuthzDecisionQuery()
        request = self._makeRequest(query=query)
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        response = self.app.post('/authorisationservice/', 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        print("Response status=%d" % response.status)
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.SUCCESS_URI)
        self.assert_(samlResponse.inResponseTo == query.id)
        self.assert_(samlResponse.assertions[0].subject.nameID.value == \
                     query.subject.nameID.value)
        self.assert_(samlResponse.assertions[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0
                                            ].decision == DecisionType.PERMIT)
   
    def test02AccessDenied(self):
        query = self._createAuthzDecisionQuery(action=Action.HTTP_POST_ACTION)
        request = self._makeRequest(query=query)
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        response = self.app.post('/authorisationservice/', 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.SUCCESS_URI)
        self.assert_(samlResponse.inResponseTo == query.id)
        self.assert_(samlResponse.assertions[0].subject.nameID.value == \
                     query.subject.nameID.value)
        self.assert_(samlResponse.assertions[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0
                                            ].decision == DecisionType.DENY)
        
    def test03IndeterminateResponse(self):
        query = self._createAuthzDecisionQuery(
                            resource=self.__class__.RESOURCE_URI + 'invalid')
        request = self._makeRequest(query=query)
        
        header = {
            'soapAction': "http://www.oasis-open.org/committees/security",
            'Content-length': str(len(request)),
            'Content-type': 'text/xml'
        }
        response = self.app.post('/authorisationservice/', 
                                 params=request, 
                                 headers=header, 
                                 status=200)
        print("Response status=%d" % response.status)
        
        samlResponse = self._getSAMLResponse(response.body)

        self.assert_(samlResponse.status.statusCode.value == \
                     StatusCode.SUCCESS_URI)
        self.assert_(samlResponse.inResponseTo == query.id)
        self.assert_(samlResponse.assertions[0].subject.nameID.value == \
                     query.subject.nameID.value)
        self.assert_(samlResponse.assertions[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0])
        self.assert_(samlResponse.assertions[0].authzDecisionStatements[0
                                    ].decision == DecisionType.INDETERMINATE)
        

if __name__ == "__main__":
    unittest.main()


"""SAML 2.0 bindings module implements SOAP binding for attribute query

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "02/09/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7259 2010-08-02 12:26:32Z pjkersha $'
import logging
log = logging.getLogger(__name__)

from os import path
from ConfigParser import ConfigParser, SafeConfigParser

from ndg.saml.common import SAMLObject

from ndg.saml.utils.factory import importModuleObject
from ndg.soap import SOAPEnvelopeBase
from ndg.soap.etree import SOAPEnvelope
from ndg.soap.client import (UrlLib2SOAPClient, UrlLib2SOAPRequest)

from ndg.saml.saml2.binding.soap import SOAPBindingInvalidResponse
    
    
_isIterable = lambda obj: getattr(obj, '__iter__', False) 
   

class SOAPBinding(object):
    '''Client SAML SOAP Binding'''
    SOAP_ACTION = 'http://www.oasis-open.org/committees/security'
    
    REQUEST_ENVELOPE_CLASS_OPTNAME = 'requestEnvelopeClass'
    RESPONSE_ENVELOPE_CLASS_OPTNAME = 'responseEnvelopeClass'
    SERIALISE_OPTNAME = 'serialise'
    DESERIALISE_OPTNAME = 'deserialise'  
    
    CONFIG_FILE_OPTNAMES = (
        REQUEST_ENVELOPE_CLASS_OPTNAME,
        RESPONSE_ENVELOPE_CLASS_OPTNAME,
        SERIALISE_OPTNAME,
        DESERIALISE_OPTNAME
    )
    
    __PRIVATE_ATTR_PREFIX = "__"
    __slots__ = tuple([__PRIVATE_ATTR_PREFIX + i 
                       for i in CONFIG_FILE_OPTNAMES + ("client",)])
    del i
    
    isIterable = staticmethod(_isIterable)
    
    def __init__(self, 
                 requestEnvelopeClass=SOAPEnvelope,
                 responseEnvelopeClass=SOAPEnvelope,
                 serialise=None,
                 deserialise=None,
                 handlers=()):
        '''Create SAML SOAP Client - Nb. serialisation functions must be set
        before send()ing the request'''
        self.__client = None
        self.__serialise = None
        self.__deserialise = None
        
        if serialise is not None:
            self.serialise = serialise
            
        if deserialise is not None:
            self.deserialise = deserialise
        
        self.client = UrlLib2SOAPClient()
        self.client.httpHeader['SOAPAction'] = SOAPBinding.SOAP_ACTION
        
        # Configurable envelope classes
        self.requestEnvelopeClass = requestEnvelopeClass
        self.client.responseEnvelopeClass = responseEnvelopeClass

        if not SOAPBinding.isIterable(handlers):
            raise TypeError('Expecting iterable for "handlers" keyword; got %r'
                            % type(handlers))
           
        for handler in handlers:
            self.client.openerDirector.add_handler(handler())

    def _getSerialise(self):
        return self.__serialise

    def _setSerialise(self, value):
        if isinstance(value, basestring):
            self.__serialise = importModuleObject(value)
            
        elif callable(value):
            self.__serialise = value
        else:
            raise TypeError('Expecting callable for "serialise"; got %r' % 
                            value)

    serialise = property(_getSerialise, _setSerialise, 
                         doc="callable to serialise request into XML type")

    def _getDeserialise(self):
        return self.__deserialise

    def _setDeserialise(self, value):
        if isinstance(value, basestring):
            self.__deserialise = importModuleObject(value)
            
        elif callable(value):
            self.__deserialise = value
        else:
            raise TypeError('Expecting callable for "deserialise"; got %r' % 
                            value)
        

    deserialise = property(_getDeserialise, 
                           _setDeserialise, 
                           doc="callable to de-serialise response from XML "
                               "type")

    def _getRequestEnvelopeClass(self):
        return self.__requestEnvelopeClass

    def _setRequestEnvelopeClass(self, value):
        if isinstance(value, basestring):
            self.client.responseEnvelopeClass = importClass(value)
            
        elif issubclass(value, SOAPEnvelopeBase):
            self.client.responseEnvelopeClass = value
        else:
            raise TypeError('Expecting %r derived type or string for '
                            '"requestEnvelopeClass" attribute; got %r' % 
                            (SOAPEnvelopeBase, value))
        
        self.__requestEnvelopeClass = value

    requestEnvelopeClass = property(_getRequestEnvelopeClass, 
                                    _setRequestEnvelopeClass, 
                                    doc="SOAP Envelope Request Class")

    def _getClient(self):
        return self.__client

    def _setClient(self, value):     
        if not isinstance(value, UrlLib2SOAPClient):
            raise TypeError('Expecting %r for "client"; got %r' % 
                            (UrlLib2SOAPClient, type(value)))
        self.__client = value

    client = property(_getClient, _setClient, 
                      doc="SOAP Client object")   

    def send(self, samlObj, uri=None, request=None):
        '''Make an request/query to a remote SAML service
        
        @type samlObj: saml.common.SAMLObject
        @param samlObj: SAML query/request object
        @type uri: basestring 
        @param uri: uri of service.  May be omitted if set from request.url
        @type request: ndg.security.common.soap.UrlLib2SOAPRequest
        @param request: SOAP request object to which query will be attached
        defaults to ndg.security.common.soap.client.UrlLib2SOAPRequest
        '''
        if self.serialise is None:
            raise AttributeError('No "serialise" method set to serialise the '
                                 'request')

        if self.deserialise is None:
            raise AttributeError('No "deserialise" method set to deserialise '
                                 'the response')
           
        if not isinstance(samlObj, SAMLObject):
            raise TypeError('Expecting %r for input attribute query; got %r'
                            % (SAMLObject, type(samlObj)))
            
        if request is None:
            request = UrlLib2SOAPRequest()
            request.envelope = self.requestEnvelopeClass()
            request.envelope.create()
            
        if uri is not None:
            request.url = uri
        
        samlElem = self.serialise(samlObj)
            
        # Attach query to SOAP body
        request.envelope.body.elem.append(samlElem)
            
        response = self.client.send(request)
        
        if len(response.envelope.body.elem) != 1:
            raise SOAPBindingInvalidResponse("Expecting single child element "
                                             "is SOAP body")
            
        response = self.deserialise(response.envelope.body.elem[0])
        
        return response

    @classmethod
    def fromConfig(cls, cfg, **kw):
        '''Alternative constructor makes object from config file settings
        @type cfg: basestring / ConfigParser derived type
        @param cfg: configuration file path or ConfigParser type object
        @rtype: ndg.saml.saml2.binding.soap.client.SOAPBinding or derived type
        @return: new instance of this class
        '''
        obj = cls()
        obj.parseConfig(cfg, **kw)
        
        return obj

    def parseConfig(self, cfg, prefix='', section='DEFAULT'):
        '''Read config file settings
        @type cfg: basestring /ConfigParser derived type
        @param cfg: configuration file path or ConfigParser type object
        @type prefix: basestring
        @param prefix: prefix for option names e.g. "attributeQuery."
        @type section: baestring
        @param section: configuration file section from which to extract
        parameters.
        '''  
        if isinstance(cfg, basestring):
            cfgFilePath = path.expandvars(cfg)
            hereDir = path.dirname(cfgFilePath)
            _cfg = SafeConfigParser(defaults=dict(here=hereDir))
            _cfg.optionxform = str

            _cfg.read(cfgFilePath)
            
        elif isinstance(cfg, ConfigParser):
            _cfg = cfg   
        else:
            raise AttributeError('Expecting basestring or ConfigParser type '
                                 'for "cfg" attribute; got %r type' % type(cfg))
        
        # Get items for this section as a dictionary so that parseKeywords can
        # used to update the object
        kw = dict(_cfg.items(section))
        if 'prefix' not in kw and prefix:
            kw['prefix'] = prefix
            
        self.parseKeywords(**kw)
        
    def parseKeywords(self, prefix='', **kw):
        """Update object from input keywords
        
        @type prefix: basestring
        @param prefix: if a prefix is given, only update self from kw items 
        where keyword starts with this prefix
        @type kw: dict
        @param kw: items corresponding to class instance variables to 
        update.  Keyword names must match their equivalent class instance 
        variable names.  However, they may prefixed with <prefix>
        """
        prefixLen = len(prefix)
        for optName, val in kw.items():
            if prefix:
                # Filter attributes based on prefix
                if optName.startswith(prefix):
                    setattr(self, optName[prefixLen:], val)
            else:
                # No prefix set - attempt to set all attributes   
                setattr(self, optName, val)
                
    @classmethod
    def fromKeywords(cls, prefix='', **kw):
        """Create a new instance initialising instance variables from the 
        keyword inputs
        @type prefix: basestring
        @param prefix: if a prefix is given, only update self from kw items 
        where keyword starts with this prefix
        @type kw: dict
        @param kw: items corresponding to class instance variables to 
        update.  Keyword names must match their equivalent class instance 
        variable names.  However, they may prefixed with <prefix>
        @return: new instance of this class
        @rtype: ndg.saml.saml2.binding.soap.client.SOAPBinding or derived type
        """
        obj = cls()
        obj.fromKeywords(prefix=prefix, **kw)
        
        return obj
        
    def __setattr__(self, name, value):
        """Enable setting of SOAPBinding.client.responseEnvelopeClass as if it
        were an attribute of self
        """
        try:
            super(SOAPBinding, self).__setattr__(name, value)
            
        except AttributeError:
            if 'name' == SOAPBinding.RESPONSE_ENVELOPE_CLASS_OPTNAME:
                if isinstance(value, basestring):
                    self.client.responseEnvelopeClass = importClass(value)
                elif issubclass(value, SOAPEnvelopeBase):
                    self.client.responseEnvelopeClass = value
                else:
                    raise TypeError('Expecting string or type instance for %r; '
                                    'got %r instead.' % (name, type(value)))
            else:
                raise
                    
    def __getstate__(self):
        '''Explicit implementation needed with __slots__'''
        _dict = {}
        for attrName in SOAPBinding.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_SOAPBinding" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
        
    def __setstate__(self, attrDict):
        '''Explicit implementation needed with __slots__'''
        for attr, val in attrDict.items():
            setattr(self, attr, val)



"""Utilities module for NDG Security SAML implementation

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "10/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7143 2010-07-01 07:53:38Z pjkersha $'
try:
    from datetime import strptime
except ImportError:
    # Allow for Python < 2.5
    from time import strptime as _strptime
    strptime = lambda datetimeStr, format: datetime(*(_strptime(datetimeStr, 
                                                                format)[0:6]))
from datetime import datetime, timedelta

        
# Interpret a string as a boolean
str2Bool = lambda str: str.lower() in ("yes", "true", "t", "1")

      
class SAMLDateTime(object):
    """Generic datetime formatting utility for SAML timestamps - XMLSchema
    Datetime format
    
    @cvar DATETIME_FORMAT: date/time format string for SAML timestamps
    @type DATETIME_FORMAT: string
    """
    DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
    
    @classmethod
    def toString(cls, dtValue):
        """Convert issue instant datetime to correct string type for output
        
        @type dtValue: datetime.datetime
        @param dtValue: issue instance as a datetime
        @rtype: basestring
        @return: issue instance as a string
        """
        if not isinstance(dtValue, datetime):
            raise TypeError("Expecting datetime type for string conversion, "
                            "got %r" % dtValue)
            
        # isoformat provides the correct formatting
#        return dtIssueInstant.strftime(cls.DATETIME_FORMAT)
        return datetime.isoformat(dtValue)+'Z'

    @classmethod
    def fromString(cls, strDateTime):
        """Convert issue instant string to datetime type
        
        @type strDateTime: basestring
        @param strDateTime: issue instance as a string
        @rtype: datetime.datetime
        @return: issue instance as a datetime
        """
        if not isinstance(strDateTime, basestring):
            raise TypeError("Expecting basestring derived type for string "
                            "conversion, got %r" % strDateTime)
        
        # Workaround for seconds fraction as strptime doesn't seem able to deal
        # with this 
        strDateTimeFraction, strSecondsFraction = strDateTime.split('.')
        dtValue = datetime.strptime(strDateTimeFraction, cls.DATETIME_FORMAT)
        secondsFraction = float("0." + strSecondsFraction.replace('Z', ''))
        dtValue += timedelta(seconds=secondsFraction)
        return dtValue


class TypedList(list):
    """Extend list type to enabled only items of a given type.  Supports
    any type where the array type in the Standard Library is restricted to 
    only limited set of primitive types
    """
    
    def __init__(self, elementType, *arg, **kw):
        """
        @type elementType: type/tuple
        @param elementType: object type or types which the list is allowed to
        contain.  If more than one type, pass as a tuple
        """
        self.__elementType = elementType
        super(TypedList, self).__init__(*arg, **kw)
    
    def _getElementType(self):
        """@return: element type for this list
        @rtype: type
        """
        return self.__elementType
    
    elementType = property(fget=_getElementType, 
                           doc="The allowed type or types for list elements")
     
    def extend(self, iter):
        """Extend an existing list with the input iterable
        @param iter: iterable to extend list with
        @type iter: iterable
        """
        for i in iter:
            if not isinstance(i, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
                
        return super(TypedList, self).extend(iter)
        
    def __iadd__(self, iter):
        """Extend an existing list with the input iterable with += operator
        
        @param iter: iterable to extend list with
        @type iter: iterable
        """
        for i in iter:
            if not isinstance(i, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
                    
        return super(TypedList, self).__iadd__(iter)
         
    def append(self, item):
        """Append a list with the given item
        
        @param item: item to extend list
        @type item: must agree witj "elementType" attribute of this list 
        """
        if not isinstance(item, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
    
        return super(TypedList, self).append(item)



"""SAML SOAP Binding Query/Response Interface with service hosted in
Paste paster web server

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "01/07/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id:$'
import logging
logging.basicConfig(level=logging.DEBUG)

import unittest
from os import path
from xml.etree import ElementTree

from ndg.soap.utils.etree import prettyPrint

from ndg.saml.saml2.core import Attribute, StatusCode
from ndg.saml.xml.etree import ResponseElementTree
from ndg.saml.saml2.binding.soap.client.attributequery import \
    AttributeQuerySslSOAPBinding
from ndg.saml.test.binding.soap import WithPasterBaseTestCase    
    
    
class SamlSslSoapBindingTestCase(WithPasterBaseTestCase):
    """Test SAML SOAP Binding with SSL"""
    SERVICE_URI = 'https://localhost:5443/attributeauthority'
    SUBJECT = "https://openid.localhost/philip.kershaw"
    SUBJECT_FORMAT = "urn:ndg:saml:openid"
    CONFIG_FILENAME = 'attribute-interface.ini'
    
    CLIENT_CERT_FILEPATH = path.join(WithPasterBaseTestCase.THIS_DIR, 
                                     'test.crt')
    CLIENT_PRIKEY_FILEPATH = path.join(WithPasterBaseTestCase.THIS_DIR, 
                                       'test.key')
    CLIENT_CACERT_DIR = path.join(WithPasterBaseTestCase.THIS_DIR, 'ca')
    VALID_DNS = [
        '/O=NDG/OU=Security/CN=localhost', 
    ]
    
    def __init__(self, *arg, **kw):
        kw['withSSL'] = True
        super(SamlSslSoapBindingTestCase, self).__init__(*arg, **kw)
                
    def test01M2CryptoInstalled(self):
        # Force error for M2Crypto not present
        _support = AttributeQuerySslSOAPBinding.SSL_CONTEXT_PROXY_SUPPORT
        AttributeQuerySslSOAPBinding.SSL_CONTEXT_PROXY_SUPPORT = False
        try:
            self.assertRaises(ImportError, AttributeQuerySslSOAPBinding)
        finally:
            AttributeQuerySslSOAPBinding.SSL_CONTEXT_PROXY_SUPPORT = _support
        
        # Try again to really test all is well
        try:
            AttributeQuerySslSOAPBinding()
        except ImportError, e:
            self.fail('Import error with AttributeQuerySslSoapBinding: %s' % e)
     
    def test02SendQuery(self):
        attributeQuery = AttributeQuerySslSOAPBinding()
        
        attributeQuery.subjectID = self.__class__.SUBJECT
        attributeQuery.subjectIdFormat = self.__class__.SUBJECT_FORMAT
        attributeQuery.clockSkewTolerance = 2.
        attributeQuery.issuerName = '/O=Site A/CN=Authorisation Service'
        
        attribute = Attribute()
        attribute.name = 'urn:ndg:saml:emailaddress'
        attribute.friendlyName = 'emailAddress'
        attribute.nameFormat = 'http://www.w3.org/2001/XMLSchema'
        
        attributeQuery.queryAttributes.append(attribute)
        
        attributeQuery.sslCACertDir = self.__class__.CLIENT_CACERT_DIR
        attributeQuery.sslCertFilePath = self.__class__.CLIENT_CERT_FILEPATH
        attributeQuery.sslPriKeyFilePath = self.__class__.CLIENT_PRIKEY_FILEPATH
        attributeQuery.sslValidDNs = self.__class__.VALID_DNS
        
        response = attributeQuery.send(uri=self.__class__.SERVICE_URI)
        
        # Convert back to ElementTree instance read for string output
        samlResponseElem = ResponseElementTree.toXML(response)
        
        print("SAML Response ...")
        print(ElementTree.tostring(samlResponseElem))
        print("Pretty print SAML Response ...")
        print(prettyPrint(samlResponseElem))
        
        self.assert_(response.status.statusCode.value==StatusCode.SUCCESS_URI)
  
 
if __name__ == "__main__":
    unittest.main()        


"""SAML 2.0 common package

Implementation of SAML 2.0 for NDG Security

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "11/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $"
from ndg.saml.common.xml import SAMLConstants, QName
 

class SAMLObject(object):
    """Base class for all SAML types
    
    @cvar DEFAULT_ELEMENT_LOCAL_NAME: default XML element name - derived classes
    must specify 
    @type DEFAULT_ELEMENT_LOCAL_NAME: None
    @ivar __qname: qualified name for XML element
    @type __qname: ndg.saml.common.xml.QName
    """
    DEFAULT_ELEMENT_LOCAL_NAME = None
    __slots__ = ('__qname',)
    
    def __init__(self,
                 namespaceURI=SAMLConstants.SAML20_NS, 
                 elementLocalName=None, 
                 namespacePrefix=SAMLConstants.SAML20_PREFIX):
        '''
        @param namespaceURI: the namespace the element is in
        @type namespaceURI: basestring
        @param elementLocalName: the local name of the XML element this Object 
        represents, defaults to DEFAULT_ELEMENT_LOCAL_NAME.  Ensure that this
        is set to a valid string in derived classes rather the None base class
        setting
        @type elementLocalName: NoneType/basestring
        @param namespacePrefix: the prefix for the given namespace
        @type namespacePrefix: basestring
        '''
        if elementLocalName is None:
            elementLocalName = self.__class__.DEFAULT_ELEMENT_LOCAL_NAME
            
        self.__qname = QName(namespaceURI, 
                             elementLocalName, 
                             namespacePrefix)
            
    @property
    def qname(self):
        """Qualified Name for this type
        
        @return: qualified name
        @rtype: ndg.saml.common.xml.QName
        """
        return self.__qname
            
    @classmethod
    def fromXML(cls, xmlObject):
        '''Parse from an XML representation into a SAML object.  Abstract method
        - derived types should implement
        
        @type xmlObject: XML class e.g. ElementTree or 4Suite XML type
        @param xmlObject: XML representation of SAML Object
        @rtype: saml.saml2.common.SAMLObject derived type
        @return: SAML object
        '''
        raise NotImplementedError()
    
    @classmethod
    def toXML(cls, samlObject):
        '''Convert the input SAML object into an XML representation.  Abstract 
        method - derived types should implement
        @type samlObject: saml.saml2.common.SAMLObject derived type
        @param samlObject: SAML object
        @rtype: XML class e.g. ElementTree or 4Suite XML
        @return: XML representation of SAML Object
        '''
        raise NotImplementedError()

    def __getstate__(self):
        '''Enable pickling
        
        @return: object's attribute dictionary
        @rtype: dict
        '''
        _dict = {}
        for attrName in SAMLObject.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_SAMLObject" + attrName
                
            try:
                _dict[attrName] = getattr(self, attrName)
            except:
                pass
            
        return _dict
  
    def __setstate__(self, attrDict):
        '''Enable pickling
        
        @param attrDict: object's attribute dictionary
        @type attrDict: dict
        '''
        for attrName, val in attrDict.items():
            setattr(self, attrName, val)
            

class SAMLVersion(object):
    """Version helper class
    
    @cvar VERSION_10: SAML Version 1.0 identifier
    @type VERSION_10: tuple
    @cvar VERSION_11: SAML Version 1.1 identifier
    @type VERSION_11: tuple
    @cvar VERSION_20: SAML Version 2.0 identifier
    @type VERSION_20: tuple
    @cvar KNOWN_VERSIONS: list of known SAML version identifiers
    @type KNOWN_VERSIONS: tuple
    @ivar __version: SAML version for the given class instance
    @type __version: tuple
    """
    
    VERSION_10 = (1, 0)
    VERSION_11 = (1, 1)
    VERSION_20 = (2, 0)
    KNOWN_VERSIONS = (VERSION_10, VERSION_11, VERSION_20)
    
    __slots__ = ('__version', )
    
    def __init__(self, version):
        """Instantiate from a given input version
        @param version: SAML version to set
        @type version: basestring or tuple or list
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, basestring):
            self.__version = SAMLVersion.valueOf(version)
        elif isinstance(version, (tuple, list)):
            self.__version = tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version initialiser; got %r" % version)
            
    def __getstate__(self):
        '''Enable pickling
        
        @return: object's attribute dictionary
        @rtype: dict
        '''
        _dict = {}
        for attrName in SAMLVersion.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_SAMLVersion" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
  
    def __setstate__(self, attrDict):
        '''Enable pickling
        
        @param attrDict: object's attribute dictionary
        @type attrDict: dict
        '''
        for attrName, val in attrDict.items():
            setattr(self, attrName, val)
    
    def __str__(self):
        """
        @return: string representation of SAML version
        @rtype: string
        """
        return ".".join([str(i) for i in self.__version])
    
    def __eq__(self, version):
        """Test for equality against an input version string, tuple or list
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if input and this object match
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, SAMLVersion):
            return str(self) == str(version)
          
        elif isinstance(version, basestring):
            return self.__version == SAMLVersion.valueOf(version)
        
        elif isinstance(version, (tuple, list)):
            return self.__version == tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version comparison; got %r" % version)
            
    def __ne__(self, version):
        """Test True for this instance version not equal to input version
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if input and this object don't match
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        return not self.__eq__(version)
            
    def __gt__(self, version):                
        """Test True for this instance version greater than input version
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if this instance version greater than input version
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, basestring):
            return self.__version > SAMLVersion.valueOf(version)
        elif isinstance(version, (tuple, list)):
            return self.__version > tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version comparison; got %r" % version)
            
    def __lt__(self, version):
        """Test True for this instance version less than input version
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if this instance version less than input version
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, basestring):
            return self.__version < SAMLVersion.valueOf(version)
        elif isinstance(version, (tuple, list)):
            return self.__version < tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version comparison; got %r" % version)
            
    def __ge__(self, version):                
        """Test True for this instance version greater or equal to the input 
        version
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if this instance version greater than or equal to input 
        version
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, basestring):
            return self.__version >= SAMLVersion.valueOf(version)
        elif isinstance(version, (tuple, list)):
            return self.__version >= tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version comparison; got %r" % version)
            
    def __le__(self, version):                
        """Test True for this instance version less than or equal to input 
        version
        
        @param version: SAML version to test
        @type version: SAMLVersion, basestring, tuple or list
        @return: True if this instance version less than or equal to input 
        version
        @rtype: bool
        @raise TypeError: unexpected type for version input
        """
        if isinstance(version, basestring):
            return self.__version <= SAMLVersion.valueOf(version)
        elif isinstance(version, (tuple, list)):
            return self.__version <= tuple(version)
        else:
            raise TypeError("Expecting string, tuple or list type for SAML "
                            "version comparison; got %r" % version)
   
    @staticmethod
    def valueOf(version):
        """Parse input string into version tuple
        @type version: basestring
        @param version: SAML version
        @rtype: tuple
        @return: SAML version tuple"""
        return tuple([int(i) for i in version.split(".")])
